"""`#include <net/if.h>`

The associated manpage is netdevice(7)

"""
from __future__ import annotations
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.sys.socket import Sockaddr
from rsyscall.struct import Struct
import enum
import typing as t

__all__ = [
    "IFF",
    "SIOC",
    "TUNSETIFF",
    "Ifreq",
]

TUNSETIFF: int = lib.TUNSETIFF

class SIOC(enum.IntEnum):
    "ioctl request codes for network device configuration"
    GIFNAME = lib.SIOCGIFNAME # get iface name
    SIFLINK = lib.SIOCSIFLINK # set iface channel
    GIFCONF = lib.SIOCGIFCONF # get iface list
    GIFFLAGS = lib.SIOCGIFFLAGS # get flags
    SIFFLAGS = lib.SIOCSIFFLAGS # set flags
    GIFADDR = lib.SIOCGIFADDR # get PA address
    SIFADDR = lib.SIOCSIFADDR # set PA address
    GIFDSTADDR = lib.SIOCGIFDSTADDR # get remote PA address
    SIFDSTADDR = lib.SIOCSIFDSTADDR # set remote PA address
    GIFBRDADDR = lib.SIOCGIFBRDADDR # get broadcast PA address
    SIFBRDADDR = lib.SIOCSIFBRDADDR # set broadcast PA address
    GIFNETMASK = lib.SIOCGIFNETMASK # get network PA mask
    SIFNETMASK = lib.SIOCSIFNETMASK # set network PA mask
    GIFMETRIC = lib.SIOCGIFMETRIC # get metric
    SIFMETRIC = lib.SIOCSIFMETRIC # set metric
    GIFMEM = lib.SIOCGIFMEM # get memory address (BSD)
    SIFMEM = lib.SIOCSIFMEM # set memory address (BSD)
    GIFMTU = lib.SIOCGIFMTU # get MTU size
    SIFMTU = lib.SIOCSIFMTU # set MTU size
    SIFNAME = lib.SIOCSIFNAME # set interface name
    SIFHWADDR = lib.SIOCSIFHWADDR # set hardware address
    GIFENCAP = lib.SIOCGIFENCAP # get/set encapsulations
    SIFENCAP = lib.SIOCSIFENCAP
    GIFHWADDR = lib.SIOCGIFHWADDR # Get hardware address
    GIFSLAVE = lib.SIOCGIFSLAVE # Driver slaving support
    SIFSLAVE = lib.SIOCSIFSLAVE
    ADDMULTI = lib.SIOCADDMULTI # Multicast address lists
    DELMULTI = lib.SIOCDELMULTI
    GIFINDEX = lib.SIOCGIFINDEX # name -> if_index mapping
    SIFPFLAGS = lib.SIOCSIFPFLAGS # set/get extended flags set
    GIFPFLAGS = lib.SIOCGIFPFLAGS
    DIFADDR = lib.SIOCDIFADDR # delete PA address
    SIFHWBROADCAST = lib.SIOCSIFHWBROADCAST # set hardware broadcast addr
    GIFCOUNT = lib.SIOCGIFCOUNT # get number of devices

class IFF(enum.IntFlag):
    """Flags which can be set in Ifreq.flags

    These flags overlap; different sets of flags are used by different
    operations.

    """
    NONE        = 0
    ## device flags, from SIOC{G,S}IFFLAGS
    UP          = lib.IFF_UP          # Interface is running.
    BROADCAST   = lib.IFF_BROADCAST   # Valid broadcast address set.
    DEBUG       = lib.IFF_DEBUG       # Internal debugging flag.
    LOOPBACK    = lib.IFF_LOOPBACK    # Interface is a loopback interface.
    POINTOPOINT = lib.IFF_POINTOPOINT # Interface is a point-to-point link.
    RUNNING     = lib.IFF_RUNNING     # Resources allocated.
    NOARP       = lib.IFF_NOARP       # No arp protocol, L2 destination address not set.
    PROMISC     = lib.IFF_PROMISC     # Interface is in promiscuous mode.
    NOTRAILERS  = lib.IFF_NOTRAILERS  # Avoid use of trailers.
    ALLMULTI    = lib.IFF_ALLMULTI    # Receive all multicast packets.
    MASTER      = lib.IFF_MASTER      # Master of a load balancing bundle.
    SLAVE       = lib.IFF_SLAVE       # Slave of a load balancing bundle.
    MULTICAST   = lib.IFF_MULTICAST   # Supports multicast
    PORTSEL     = lib.IFF_PORTSEL     # Is able to select media type via ifmap.
    AUTOMEDIA   = lib.IFF_AUTOMEDIA   # Auto media selection active.
    DYNAMIC     = lib.IFF_DYNAMIC     # The addresses are lost when the interface goes down.
    # glibc doesn't have these in its headers for some reason
    # LOWER_UP    = lib.IFF_LOWER_UP    # Driver signals L1 up (since Linux 2.6.17)
    # DORMANT     = lib.IFF_DORMANT     # Driver signals dormant (since Linux 2.6.17)
    # ECHO        = lib.IFF_ECHO        # Echo sent packets (since Linux 2.6.25)
    ## TUNSETIFF flags
    TUN         = lib.IFF_TUN
    TAP         = lib.IFF_TAP
    NO_PI       = lib.IFF_NO_PI

class CStringField:
    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(self, instance, owner) -> str:
        data = bytes(ffi.buffer(getattr(instance.cffi, self.name)))
        try:
            valid_data = data[:data.index(b'\0')]
        except ValueError:
            valid_data = data
        return valid_data.decode()

    def __set__(self, instance, value: str) -> None:
        setattr(instance.cffi, self.name, value.encode() + b"\0")

class IntField:
    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(self, instance, owner) -> int:
        return getattr(instance.cffi, self.name)

    def __set__(self, instance, value: int) -> None:
        setattr(instance.cffi, self.name, value)

class AddressField:
    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(self, instance, owner) -> Sockaddr:
        data_bytes = bytes(ffi.buffer(ffi.addressof(instance.cffi, self.name)))
        return Sockaddr.from_bytes(data_bytes)

    def __set__(self, instance, value: Sockaddr) -> None:
        data_bytes = value.to_bytes()
        ffi.memmove(ffi.addressof(instance.cffi, self.name),
                    ffi.from_buffer(data_bytes), len(data_bytes))

class IFFField:
    def __init__(self, name: str) -> None:
        self.name = name

    def __get__(self, instance, owner) -> IFF:
        return IFF(getattr(instance.cffi, self.name))

    def __set__(self, instance, value: IFF) -> None:
        setattr(instance.cffi, self.name, value)

class Ifreq(Struct):
    """Representation of "struct ifreq"

    We have to be somewhat careful in how we represent this, since this struct
    is one big union, plus ifr_name which can be unset anyway.

    The way we handle this is, all the fields on this class are properties which
    extract some specific field from the union, stored as a cffi type.

    """
    name = CStringField("ifr_name")
    addr = AddressField("ifr_addr")
    ifindex = IntField("ifr_ifindex")
    flags = IFFField("ifr_flags")
    
    def __init__(self, name: str=None, *, addr: Sockaddr=None, flags: IFF=None, cffi=None) -> None:
        if cffi is None:
            cffi = ffi.new('struct ifreq*')
        self.cffi = cffi
        if name is not None:
            self.name = name
        if addr is not None:
            self.addr = addr
        if flags is not None:
            self.flags = flags

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(self.cffi))

    T = t.TypeVar('T', bound='Ifreq')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        if len(data) != cls.sizeof():
            raise Exception("data length", len(data),
                            "doesn't match actual length of struct ifreq", cls.sizeof())
        cffi = ffi.new('struct ifreq*')
        ffi.memmove(cffi, ffi.from_buffer(data), cls.sizeof())
        return cls(cffi=cffi)

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct ifreq')

    def __str__(self) -> str:
        if self.name:
            return f"Ifreq({self.name}, ...)"
        else:
            # if it's an empty string, indicate that there's no name
            return "Ifreq(<no interface name>, ...)"


#### Tests ####
from unittest import TestCase
from rsyscall.netinet.ip import SockaddrIn
class TestIf(TestCase):
    def test_ifreq(self) -> None:
        initial = Ifreq()
        initial.name = "hello"
        initial.addr = SockaddrIn(42, "127.0.0.1")
        output = Ifreq.from_bytes(initial.to_bytes())
        self.assertEqual(initial.name, output.name)
        addr = t.cast(SockaddrIn, initial.addr.parse())
        self.assertIsInstance(addr, SockaddrIn)
        self.assertEqual(initial.addr.port, addr.port)
        self.assertEqual(initial.addr.addr, addr.addr)



"`#include <sys/signalfd.h>`"
import typing as t
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.struct import Struct
from dataclasses import dataclass

from rsyscall.signal import SIG, Sigset
import enum

class SFD(enum.IntFlag):
    NONE = 0
    NONBLOCK = lib.SFD_NONBLOCK
    CLOEXEC = lib.SFD_CLOEXEC

@dataclass
class SignalfdSiginfo(Struct):
    # TODO fill in the rest of the data
    # (even though we don't use any of it ourselves)
    signo: SIG

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct signalfd_siginfo*')
        struct.ssi_signo = self.signo
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='SignalfdSiginfo')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct signalfd_siginfo const*', ffi.from_buffer(data))
        return cls(
            signo=SIG(struct.ssi_signo),
        )

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct signalfd_siginfo')

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask
from rsyscall.handle.pointer import Pointer

T_fd = t.TypeVar('T_fd', bound='SignalFileDescriptor')
class SignalFileDescriptor(BaseFileDescriptor):
    async def signalfd(self, mask: Pointer[Sigset], flags: SFD) -> None:
        self._validate()
        with mask.borrow(self.task) as mask_n:
            await _signalfd(self.task.sysif, self.near, mask_n, mask.size(), flags)

class SignalfdTask(FileDescriptorTask[T_fd]):
    async def signalfd(self, mask: Pointer[Sigset], flags: SFD=SFD.NONE) -> T_fd:
        with mask.borrow(self) as mask_n:
            fd = await _signalfd(self.sysif, None, mask_n, mask.size(), flags|SFD.CLOEXEC)
            return self.make_fd_handle(fd)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _signalfd(sysif: SyscallInterface, fd: t.Optional[near.FileDescriptor],
                    mask: near.Address, sizemask: int, flags: SFD) -> near.FileDescriptor:
    if fd is None:
        fd = -1 # type: ignore
    return near.FileDescriptor(await sysif.syscall(SYS.signalfd4, fd, mask, sizemask, flags))



"`#include <sys/socket.h>`"
from __future__ import annotations
from rsyscall._raw import ffi, lib # type: ignore
from dataclasses import dataclass
import typing as t
import enum
from rsyscall.struct import Struct, FixedSerializer, Serializable, Serializer, HasSerializer, FixedSize
import contextlib
import abc
import struct
import rsyscall.near.types as near
from rsyscall.handle.pointer import Pointer, WrittenPointer, ReadablePointer, LinearPointer
if t.TYPE_CHECKING:
    from rsyscall.handle import FileDescriptor, Task
else:
    FileDescriptor = object
from rsyscall.sys.uio import IovecList

__all__ = [
    "AF",
    "SHUT",
    "Sockaddr",
    "SockaddrStorage",
    "SOCK",
    "SOL",
    "SO",
    "MSG",
    "Socketpair",
    "Sockbuf",
    "CmsgSCMRights",
    "CmsgList",
    "SendMsghdr",
    "RecvMsghdr",
    "RecvMsghdrOut",
    "SendmsgFlags",
    "RecvmsgFlags",
    "MsghdrFlags",
    "SocketTask",
    "SocketFileDescriptor",
]

T = t.TypeVar('T')
class AF(enum.IntEnum):
    UNIX = lib.AF_UNIX
    NETLINK = lib.AF_NETLINK
    INET = lib.AF_INET
    INET6 = lib.AF_INET6

class SHUT(enum.IntEnum):
    RD = lib.SHUT_RD
    WR = lib.SHUT_WR
    RDWR = lib.SHUT_RDWR

class MSG(enum.IntFlag):
    NONE = 0
    # send flags
    CONFIRM = lib.MSG_CONFIRM
    DONTROUTE = lib.MSG_DONTROUTE
    EOR = lib.MSG_EOR
    MORE = lib.MSG_MORE
    NOSIGNAL = lib.MSG_NOSIGNAL
    # recv flags
    CMSG_CLOEXEC = lib.MSG_CMSG_CLOEXEC
    ERRQUEUE = lib.MSG_ERRQUEUE
    PEEK = lib.MSG_PEEK
    TRUNC = lib.MSG_TRUNC
    WAITALL = lib.MSG_WAITALL
    # both
    DONTWAIT = lib.MSG_DONTWAIT
    OOB = lib.MSG_OOB

family_to_class: t.Dict[AF, t.Type[Sockaddr]] = {}
def _register_sockaddr(sockaddr: t.Type[Sockaddr]) -> None:
    if sockaddr.family in family_to_class:
        raise Exception("tried to register sockaddr", sockaddr, "for family", sockaddr.family,
                        "but there's already a class registered for that family:",
                        family_to_class[sockaddr.family])
    family_to_class[sockaddr.family] = sockaddr

class Sockaddr(Struct):
    """struct sockaddr. This is not really useful to allocate on its own; you want the derived classes.

    """
    def __init__(self, family: AF, data: bytes) -> None:
        self.family = family
        self.data = data

    @classmethod
    def check_family(cls, family: AF) -> None:
        if cls.family != family:
            raise Exception("sa_family should be", cls.family, "is instead", family)

    def to_bytes(self) -> bytes:
        struct = ffi.new('struct sockaddr*', (self.family, self.data))
        return bytes(ffi.buffer(struct))

    T = t.TypeVar('T', bound='Sockaddr')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        struct = ffi.cast('struct sockaddr*', ffi.from_buffer(data))
        return cls(struct.sa_family, bytes(ffi.buffer(struct.sa_data)))

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct sockaddr')

    def parse(self) -> Sockaddr:
        "Using the family field, return the correct Sockaddr type that this actually contains."
        cls = family_to_class[self.family]
        return cls.from_bytes(self.to_bytes())

T_sockaddr = t.TypeVar('T_sockaddr', bound='Sockaddr')

class SOCK(enum.IntFlag):
    NONE = 0
    # socket kinds
    DGRAM = lib.SOCK_DGRAM
    STREAM = lib.SOCK_STREAM
    SEQPACKET = lib.SOCK_SEQPACKET
    RAW = lib.SOCK_RAW
    # flags that can be or'd in
    CLOEXEC = lib.SOCK_CLOEXEC
    NONBLOCK = lib.SOCK_NONBLOCK

class SOL(enum.IntEnum):
    """Stands for Sock Opt Level

    This is what should be passed as the "level" argument to
    getsockopt/setsockopt.

    """
    SOCKET = lib.SOL_SOCKET
    IP = lib.SOL_IP

class SO(enum.IntEnum):
    ACCEPTCONN = lib.SO_ACCEPTCONN
    ATTACH_FILTER = lib.SO_ATTACH_FILTER
    ATTACH_BPF = lib.SO_ATTACH_BPF
    ATTACH_REUSEPORT_CBPF = lib.SO_ATTACH_REUSEPORT_CBPF
    ATTACH_REUSEPORT_EBPF = lib.SO_ATTACH_REUSEPORT_EBPF
    BINDTODEVICE = lib.SO_BINDTODEVICE
    BROADCAST = lib.SO_BROADCAST
    BSDCOMPAT = lib.SO_BSDCOMPAT
    DEBUG = lib.SO_DEBUG
    DETACH_FILTER = lib.SO_DETACH_FILTER
    DETACH_BPF = lib.SO_DETACH_BPF
    DOMAIN = lib.SO_DOMAIN
    ERROR = lib.SO_ERROR
    DONTROUTE = lib.SO_DONTROUTE
    INCOMING_CPU = lib.SO_INCOMING_CPU
    KEEPALIVE = lib.SO_KEEPALIVE
    LINGER = lib.SO_LINGER
    LOCK_FILTER = lib.SO_LOCK_FILTER
    MARK = lib.SO_MARK
    OOBINLINE = lib.SO_OOBINLINE
    PASSCRED = lib.SO_PASSCRED
    PEEK_OFF = lib.SO_PEEK_OFF
    PEERCRED = lib.SO_PEERCRED
    PRIORITY = lib.SO_PRIORITY
    PROTOCOL = lib.SO_PROTOCOL
    RCVBUF = lib.SO_RCVBUF
    RCVBUFFORCE = lib.SO_RCVBUFFORCE
    RCVLOWAT = lib.SO_RCVLOWAT
    SNDLOWAT = lib.SO_SNDLOWAT
    RCVTIMEO = lib.SO_RCVTIMEO
    SNDTIMEO = lib.SO_SNDTIMEO
    REUSEADDR = lib.SO_REUSEADDR
    REUSEPORT = lib.SO_REUSEPORT
    RXQ_OVFL = lib.SO_RXQ_OVFL
    SNDBUF = lib.SO_SNDBUF
    SNDBUFFORCE = lib.SO_SNDBUFFORCE
    TIMESTAMP = lib.SO_TIMESTAMP
    TYPE = lib.SO_TYPE
    BUSY_POLL = lib.SO_BUSY_POLL

class SCM(enum.IntEnum):
    RIGHTS = lib.SCM_RIGHTS

@dataclass
class SockaddrStorage(Sockaddr):
    "struct sockaddr_storage. Useful when dealing with sockets with unknown address families"
    family: AF
    data: bytes

    def to_bytes(self) -> bytes:
        # We can't just create a struct sockaddr_storage and turn the whole thing to bytes,
        # because that will pad the actually valid data with a bunch of trailing null bytes.
        # And we can't do that because the length of the valid data is semantically
        # meaningful for some socket addresses, such as sockaddr_un.
        return bytes(ffi.buffer(ffi.new('sa_family_t*', self.family))) + self.data

    T = t.TypeVar('T', bound='SockaddrStorage')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        # As with to_bytes, we can't just cast the bytes to a struct sockaddr_storage and read its data field,
        # because that would pad the data with a bunch of null bytes,
        # and would not preserve the length of the valid data
        family = ffi.cast('sa_family_t*', ffi.from_buffer(data))
        rest = data[ffi.sizeof('sa_family_t'):]
        return cls(
            family=AF(family[0]),
            data=rest
        )

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct sockaddr_storage')

@dataclass
class Sockbuf(t.Generic[T], HasSerializer):
    buf: Pointer[T]
    buf_rest: t.Optional[Pointer[T]] = None

    def get_self_serializer(self, task: Task) -> SockbufSerializer[T]:
        return SockbufSerializer(self.buf)

class SockbufSerializer(t.Generic[T], Serializer[Sockbuf[T]]):
    def __init__(self, buf: Pointer[T]) -> None:
        self.buf = buf

    def to_bytes(self, val: Sockbuf) -> bytes:
        return bytes(ffi.buffer(ffi.new('socklen_t*', val.buf.size())))

    def from_bytes(self, data: bytes) -> Sockbuf[T]:
        struct = ffi.cast('socklen_t*', ffi.from_buffer(data))
        socklen = struct[0]
        if socklen > self.buf.size():
            raise Exception("not enough buffer space to read socket, need", socklen)
        valid, rest = self.buf.readable_split(socklen)
        return Sockbuf(valid, rest)


#### socketpair stuff

T_socketpair = t.TypeVar('T_socketpair', bound='Socketpair')
@dataclass
class Socketpair(FixedSize):
    first: FileDescriptor
    second: FileDescriptor

    @classmethod
    def sizeof(cls) -> int:
        return ffi.sizeof('struct fdpair')

    @classmethod
    def get_serializer(cls: t.Type[T_socketpair], task: Task) -> Serializer[T_socketpair]:
        return SocketpairSerializer(cls, task)

@dataclass
class SocketpairSerializer(Serializer[T_socketpair]):
    cls: t.Type[T_socketpair]
    task: Task

    def to_bytes(self, pair: T_socketpair) -> bytes:
        struct = ffi.new('struct fdpair*', (pair.first, pair.second))
        return bytes(ffi.buffer(struct))

    def from_bytes(self, data: bytes) -> T_socketpair:
        struct = ffi.cast('struct fdpair const*', ffi.from_buffer(data))
        def make(n: int) -> FileDescriptor:
            return self.task.make_fd_handle(near.FileDescriptor(int(n)))
        return self.cls(make(struct.first), make(struct.second))


#### sendmsg/recvmsg stuff

class SendmsgFlags(enum.IntFlag):
    NONE = 0

class RecvmsgFlags(enum.IntFlag):
    NONE = 0
    CMSG_CLOEXEC = lib.MSG_CMSG_CLOEXEC

class MsghdrFlags(enum.IntFlag):
    NONE = 0
    CTRUNC = lib.MSG_CTRUNC
    CMSG_CLOEXEC = lib.MSG_CMSG_CLOEXEC

T_cmsg = t.TypeVar('T_cmsg', bound='Cmsg')
class Cmsg(FixedSerializer):
    @abc.abstractmethod
    def to_data(self) -> bytes: ...
    @abc.abstractmethod
    def borrow_with(self, stack: contextlib.ExitStack, task: FileDescriptorTask) -> None: ...
    @classmethod
    @abc.abstractmethod
    def from_data(cls: t.Type[T], task: Task, data: bytes) -> T: ...
    @classmethod
    @abc.abstractmethod
    def level(cls: t.Type[T]) -> SOL: ...
    @classmethod
    @abc.abstractmethod
    def type(cls: t.Type[T]) -> int: ...

    @classmethod
    def get_serializer(cls: t.Type[T_cmsg], task: Task) -> Serializer[T_cmsg]:
        return CmsgSerializer(cls, task)

class CmsgSerializer(Serializer[T_cmsg]):
    def __init__(self, cls: t.Type[T_cmsg], task: Task) -> None:
        self.cls = cls
        self.task = task

    def to_bytes(self, val: T_cmsg) -> bytes:
        if not isinstance(val, self.cls):
            raise Exception("Serializer for", self.cls,
                            "had to_bytes called on different type", val)
        data = val.to_data()
        header = bytes(ffi.buffer(ffi.new('struct cmsghdr*', {
            "cmsg_len": ffi.sizeof('struct cmsghdr') + len(data),
            "cmsg_level": val.level(),
            "cmsg_type": val.type(),
        })))
        return header + data

    def from_bytes(self, data: bytes) -> T_cmsg:
        record = ffi.cast('struct cmsghdr*', ffi.from_buffer(data))
        if record.cmsg_level != self.cls.level():
            raise Exception("serializer for level", self.cls.level(),
                            "got message for level", record.cmsg_level)
        if record.cmsg_type != self.cls.type():
            raise Exception("serializer for type", self.cls.type(),
                            "got message for type", record.cmsg_type)
        return self.cls.from_data(self.task, data[ffi.sizeof('struct cmsghdr'):record.cmsg_len])

import array
class CmsgSCMRights(Cmsg, t.List[FileDescriptor]):
    def to_data(self) -> bytes:
        return array.array('i', (int(fd) for fd in self)).tobytes()
    def borrow_with(self, stack: contextlib.ExitStack, task: FileDescriptorTask) -> None:
        for fd in self:
            stack.enter_context(fd.borrow(task))

    T = t.TypeVar('T', bound='CmsgSCMRights')
    @classmethod
    def from_data(cls: t.Type[T], task: Task, data: bytes) -> T:
        fds = [near.FileDescriptor(fd) for fd, in struct.Struct('i').iter_unpack(data)]
        return cls([task.make_fd_handle(fd) for fd in fds])

    @classmethod
    def level(cls) -> SOL:
        return SOL.SOCKET
    @classmethod
    def type(cls) -> int:
        return SCM.RIGHTS

T_cmsglist = t.TypeVar('T_cmsglist', bound='CmsgList')
class CmsgList(t.List[Cmsg], FixedSerializer):
    @classmethod
    def get_serializer(cls: t.Type[T_cmsglist], task: Task) -> Serializer[T_cmsglist]:
        return CmsgListSerializer(cls, task)

    def borrow_with(self, stack: contextlib.ExitStack, task: FileDescriptorTask) -> None:
        for cmsg in self:
            cmsg.borrow_with(stack, task)

class CmsgListSerializer(Serializer[T_cmsglist]):
    def __init__(self, cls: t.Type[T_cmsglist], task: Task) -> None:
        self.cls = cls
        self.task = task

    def to_bytes(self, val: T_cmsglist) -> bytes:
        ret = b""
        for cmsg in val:
            # TODO is this correct alignment/padding???
            # I don't think so...
            ret += cmsg.get_serializer(self.task).to_bytes(cmsg)
        return ret

    def from_bytes(self, data: bytes) -> T_cmsglist:
        entries = []
        while len(data) > 0:
            record = ffi.cast('struct cmsghdr*', ffi.from_buffer(data))
            record_data = data[:record.cmsg_len]
            level = SOL(record.cmsg_level)
            if level == SOL.SOCKET and record.cmsg_type == int(SCM.RIGHTS):
                entries.append(CmsgSCMRights.get_serializer(self.task).from_bytes(record_data))
            else:
                raise Exception("unknown cmsg level/type sorry", level, type)
            data = data[record.cmsg_len:]
        return self.cls(entries)

@dataclass
class SendMsghdr(Serializable):
    name: t.Optional[WrittenPointer[Sockaddr]]
    iov: WrittenPointer[IovecList]
    control: t.Optional[WrittenPointer[CmsgList]]

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct msghdr*', {
            "msg_name": ffi.cast('void*', int(self.name.near)) if self.name else ffi.NULL,
            "msg_namelen": self.name.size() if self.name else 0,
            "msg_iov": ffi.cast('void*', int(self.iov.near)),
            "msg_iovlen": len(self.iov.value),
            "msg_control": ffi.cast('void*', int(self.control.near)) if self.control else ffi.NULL,
            "msg_controllen": self.control.size() if self.control else 0,
            "msg_flags": 0,
        })))

    T = t.TypeVar('T', bound='SendMsghdr')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        raise Exception("can't get pointer handles from raw bytes")

@dataclass
class RecvMsghdr(Serializable):
    name: t.Optional[Pointer[Sockaddr]]
    iov: WrittenPointer[IovecList]
    control: t.Optional[Pointer[CmsgList]]

    def to_bytes(self) -> bytes:
        return bytes(ffi.buffer(ffi.new('struct msghdr*', {
            "msg_name": ffi.cast('void*', int(self.name.near)) if self.name else ffi.NULL,
            "msg_namelen": self.name.size() if self.name else 0,
            "msg_iov": ffi.cast('void*', int(self.iov.near)),
            "msg_iovlen": len(self.iov.value),
            "msg_control": ffi.cast('void*', int(self.control.near)) if self.control else ffi.NULL,
            "msg_controllen": self.control.size() if self.control else 0,
            "msg_flags": 0,
        })))

    T = t.TypeVar('T', bound='RecvMsghdr')
    @classmethod
    def from_bytes(cls: t.Type[T], data: bytes) -> T:
        raise Exception("can't get pointer handles from raw bytes")

    def to_out(self, ptr: Pointer[RecvMsghdr]) -> Pointer[RecvMsghdrOut]:
        # what a mouthful
        serializer = RecvMsghdrOutSerializer(self.name, self.control)
        return ptr._reinterpret(serializer, RecvMsghdrOut)

@dataclass
class RecvMsghdrOut:
    name: t.Optional[ReadablePointer[Sockaddr]]
    control: t.Optional[LinearPointer[CmsgList]]
    flags: MsghdrFlags
    # the _rest fields are the invalid, unused parts of the buffers;
    # almost everyone can ignore these.
    name_rest: t.Optional[Pointer[Sockaddr]]
    control_rest: t.Optional[Pointer[CmsgList]]

@dataclass
class RecvMsghdrOutSerializer(Serializer[RecvMsghdrOut]):
    name: t.Optional[Pointer[Sockaddr]]
    control: t.Optional[Pointer[CmsgList]]

    def to_bytes(self, x: RecvMsghdrOut) -> bytes:
        raise Exception("not going to bother implementing this")

    def from_bytes(self, data: bytes) -> RecvMsghdrOut:
        struct = ffi.cast('struct msghdr*', ffi.from_buffer(data))
        if self.name is None:
            name: t.Optional[ReadablePointer[Sockaddr]] = None
            name_rest: t.Optional[Pointer[Sockaddr]] = None
        else:
            name, name_rest = self.name.readable_split(struct.msg_namelen)
        if self.control is None:
            control: t.Optional[Pointer[CmsgList]] = None
            control_rest: t.Optional[Pointer[CmsgList]] = None
        else:
            control, control_rest = self.control.readable_split(struct.msg_controllen)
        flags = MsghdrFlags(struct.msg_flags)
        return RecvMsghdrOut(name, control._linearize() if control else None, flags, name_rest, control_rest)

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask

T_fd = t.TypeVar('T_fd', bound='SocketFileDescriptor')
class SocketFileDescriptor(BaseFileDescriptor):
    async def bind(self, addr: WrittenPointer[Sockaddr]) -> None:
        """bind a name to a socket

        man: bind(2)
        """
        self._validate()
        with addr.borrow(self.task):
            try:
                await _bind(self.task.sysif, self.near, addr.near, addr.size())
            except PermissionError as exn:
                exn.filename = addr.value
                raise
            except FileNotFoundError as exn:
                exn.filename = addr.value
                raise

    async def connect(self, addr: WrittenPointer[Sockaddr]) -> None:
        self._validate()
        with addr.borrow(self.task):
            try:
                await _connect(self.task.sysif, self.near, addr.near, addr.size())
            except OSError as exn:
                exn.filename = self
                if hasattr(addr, 'value'):
                    exn.filename2 = addr.value
                raise

    async def listen(self, backlog: int) -> None:
        self._validate()
        await _listen(self.task.sysif, self.near, backlog)

    async def getsockopt(self, level: int, optname: int, optval: WrittenPointer[Sockbuf[T]]) -> Pointer[Sockbuf[T]]:
        self._validate()
        with optval.borrow(self.task):
            with optval.value.buf.borrow(self.task):
                await _getsockopt(self.task.sysif, self.near,
                                  level, optname, optval.value.buf.near, optval.near)
        return optval

    async def setsockopt(self, level: int, optname: int, optval: Pointer) -> None:
        self._validate()
        with optval.borrow(self.task) as optval_n:
            await _setsockopt(self.task.sysif, self.near, level, optname, optval_n, optval.size())

    async def getsockname(self, addr: WrittenPointer[Sockbuf[T_sockaddr]]) -> Pointer[Sockbuf[T_sockaddr]]:
        self._validate()
        with addr.borrow(self.task) as addr_n:
            with addr.value.buf.borrow(self.task) as addrbuf_n:
                await _getsockname(self.task.sysif, self.near, addrbuf_n, addr_n)
        return addr

    async def getpeername(self, addr: WrittenPointer[Sockbuf[T_sockaddr]]) -> Pointer[Sockbuf[T_sockaddr]]:
        self._validate()
        with addr.borrow(self.task) as addr_n:
            with addr.value.buf.borrow(self.task) as addrbuf_n:
                await _getpeername(self.task.sysif, self.near, addrbuf_n, addr_n)
        return addr

    @t.overload
    async def accept(self: T_fd, flags: SOCK=SOCK.NONE) -> T_fd: ...
    @t.overload
    async def accept(self: T_fd, flags: SOCK, addr: WrittenPointer[Sockbuf[T_sockaddr]]
    ) -> t.Tuple[T_fd, WrittenPointer[Sockbuf[T_sockaddr]]]: ...

    async def accept(self: T_fd, flags: SOCK=SOCK.NONE,
                     addr: t.Optional[WrittenPointer[Sockbuf[T_sockaddr]]]=None
    ) -> t.Union[T_fd, t.Tuple[T_fd, WrittenPointer[Sockbuf[T_sockaddr]]]]:
        self._validate()
        flags |= SOCK.CLOEXEC
        if addr is None:
            fd = await _accept(self.task.sysif, self.near, None, None, flags)
            return self.task.make_fd_handle(fd)
        else:
            with addr.borrow(self.task):
                with addr.value.buf.borrow(self.task):
                    fd = await _accept(self.task.sysif, self.near,
                                       addr.value.buf.near, addr.near, flags)
                    return self.task.make_fd_handle(fd), addr

    async def shutdown(self, how: SHUT) -> None:
        self._validate()
        await _shutdown(self.task.sysif, self.near, how)

    async def sendmsg(self, msg: WrittenPointer[SendMsghdr], flags: SendmsgFlags=SendmsgFlags.NONE
    ) -> t.Tuple[IovecList, IovecList]:
        """send a message on a socket

        man: sendmsg(2)
        """
        with contextlib.ExitStack() as stack:
            stack.enter_context(msg.borrow(self.task))
            if msg.value.name:
                stack.enter_context(msg.value.name.borrow(self.task))
            if msg.value.control:
                stack.enter_context(msg.value.control.borrow(self.task))
                msg.value.control.value.borrow_with(stack, self.task)
            stack.enter_context(msg.value.iov.borrow(self.task))
            for iovec_elem in msg.value.iov.value:
                stack.enter_context(iovec_elem.borrow(self.task))
            ret = await _sendmsg(self.task.sysif, self.near, msg.near, flags)
        return msg.value.iov.value.split(ret)

    async def recvmsg(self, msg: WrittenPointer[RecvMsghdr], flags: RecvmsgFlags=RecvmsgFlags.NONE,
    ) -> t.Tuple[IovecList, IovecList, Pointer[RecvMsghdrOut]]:
        """receive a message from a socket

        man: recvmsg(2)
        """
        flags |= RecvmsgFlags.CMSG_CLOEXEC
        with contextlib.ExitStack() as stack:
            stack.enter_context(msg.borrow(self.task))
            if msg.value.name:
                stack.enter_context(msg.value.name.borrow(self.task))
            if msg.value.control:
                stack.enter_context(msg.value.control.borrow(self.task))
            stack.enter_context(msg.value.iov.borrow(self.task))
            for iovec_elem in msg.value.iov.value:
                stack.enter_context(iovec_elem.borrow(self.task))
            ret = await _recvmsg(self.task.sysif, self.near, msg.near, flags)
        valid, invalid = msg.value.iov.value.split(ret)
        return valid, invalid, msg.value.to_out(msg)

    async def recv(self, buf: Pointer[T], flags: MSG) -> t.Tuple[ReadablePointer[T], Pointer]:
        """receive a message from a socket

        man: recv(2)
        """
        self._validate()
        with buf.borrow(self.task) as buf_n:
            ret = await _recv(self.task.sysif, self.near, buf_n, buf.size(), flags)
            return buf.readable_split(ret)

    async def send(self, buf: Pointer[T], flags: MSG) -> t.Tuple[Pointer[T], Pointer]:
        """send a message on a socket

        man: send(2)
        """
        self._validate()
        with buf.borrow(self.task) as buf_n:
            ret = await _send(self.task.sysif, self.near, buf_n, buf.size(), flags)
            return buf.split(ret)

class SocketTask(FileDescriptorTask[T_fd]):
    async def socket(self, domain: AF, type: SOCK, protocol: int=0) -> T_fd:
        """create an endpoint for communication

        manpage: socket(2)
        """
        sockfd = await _socket(self.sysif, domain, type|SOCK.CLOEXEC, protocol)
        return self.make_fd_handle(sockfd)

    async def socketpair(self, domain: AF, type: SOCK, protocol: int,
                         sv: Pointer[Socketpair]) -> LinearPointer[Socketpair]:
        with sv.borrow(self) as sv_n:
            await _socketpair(self.sysif, domain, type|SOCK.CLOEXEC, protocol, sv_n)
            return sv._linearize()

#### Raw syscalls ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _socket(sysif: SyscallInterface,
                  domain: AF, type: SOCK, protocol: int) -> near.FileDescriptor:
    return near.FileDescriptor(await sysif.syscall(SYS.socket, domain, type, protocol))

async def _socketpair(sysif: SyscallInterface,
                      domain: AF, type: SOCK, protocol: int, sv: near.Address) -> None:
    await sysif.syscall(SYS.socketpair, domain, type, protocol, sv)

async def _bind(sysif: SyscallInterface, sockfd: near.FileDescriptor,
               addr: near.Address, addrlen: int) -> None:
    await sysif.syscall(SYS.bind, sockfd, addr, addrlen)

async def _connect(sysif: SyscallInterface, sockfd: near.FileDescriptor,
                   addr: near.Address, addrlen: int) -> None:
    await sysif.syscall(SYS.connect, sockfd, addr, addrlen)

async def _listen(sysif: SyscallInterface, sockfd: near.FileDescriptor,
                  backlog: int) -> None:
    await sysif.syscall(SYS.listen, sockfd, backlog)

async def _accept(sysif: SyscallInterface, sockfd: near.FileDescriptor,
                  addr: t.Optional[near.Address], addrlen: t.Optional[near.Address],
                  flags: SOCK) -> near.FileDescriptor:
    if addr is None:
        addr = 0 # type: ignore
    if addrlen is None:
        addrlen = 0 # type: ignore
    return near.FileDescriptor(await sysif.syscall(SYS.accept4, sockfd, addr, addrlen, flags))

async def _shutdown(sysif: SyscallInterface, sockfd: near.FileDescriptor, how: SHUT) -> None:
    await sysif.syscall(SYS.shutdown, sockfd, how)

async def _getsockname(sysif: SyscallInterface, sockfd: near.FileDescriptor,
                       addr: near.Address, addrlen: near.Address) -> None:
    await sysif.syscall(SYS.getsockname, sockfd, addr, addrlen)

async def _getpeername(sysif: SyscallInterface, sockfd: near.FileDescriptor,
                       addr: near.Address, addrlen: near.Address) -> None:
    await sysif.syscall(SYS.getpeername, sockfd, addr, addrlen)

async def _getsockopt(sysif: SyscallInterface, sockfd: near.FileDescriptor,
                      level: int, optname: int, optval: near.Address, optlen: near.Address) -> None:
    await sysif.syscall(SYS.getsockopt, sockfd, level, optname, optval, optlen)

async def _setsockopt(sysif: SyscallInterface, sockfd: near.FileDescriptor,
                      level: int, optname: int, optval: near.Address, optlen: int) -> None:
    await sysif.syscall(SYS.setsockopt, sockfd, level, optname, optval, optlen)

async def _recv(sysif: SyscallInterface, fd: near.FileDescriptor,
                buf: near.Address, count: int, flags: int) -> int:
    return (await sysif.syscall(SYS.recvfrom, fd, buf, count, flags))

async def _send(sysif: SyscallInterface, fd: near.FileDescriptor,
                buf: near.Address, count: int, flags: int) -> int:
    return (await sysif.syscall(SYS.sendto, fd, buf, count, flags))

async def _recvmsg(sysif: SyscallInterface, fd: near.FileDescriptor,
                   msg: near.Address, flags: int) -> int:
    return (await sysif.syscall(SYS.recvmsg, fd, msg, flags))

async def _sendmsg(sysif: SyscallInterface, fd: near.FileDescriptor,
                   msg: near.Address, flags: int) -> int:
    return (await sysif.syscall(SYS.sendmsg, fd, msg, flags))



#### Tests ####
from unittest import TestCase
class TestSocket(TestCase):
    pass



"`#include <fnctl.h>`"
from rsyscall._raw import lib, ffi # type: ignore
import enum
import typing as t

class AT(enum.IntFlag):
    """The flags argument to many *at syscall; mostly specifies changes to path resolution.

    Except for `AT.REMOVEDIR`, these are all path resolution changes.

    """
    NONE = 0
    FDCWD = lib.AT_FDCWD
    REMOVEDIR = lib.AT_REMOVEDIR
    "When passed to `FileDescriptor.unlinkat`, remove directories instead of linking files."
    EMPTY_PATH = lib.AT_EMPTY_PATH
    SYMLINK_NOFOLLOW = lib.AT_SYMLINK_NOFOLLOW
    SYMLINK_FOLLOW = lib.AT_SYMLINK_FOLLOW

class O(enum.IntFlag):
    "The flags argument to open and some other syscalls."
    NONE = 0
    RDONLY = lib.O_RDONLY
    WRONLY = lib.O_WRONLY
    RDWR = lib.O_RDWR
    CREAT = lib.O_CREAT
    EXCL = lib.O_EXCL
    NOCTTY = lib.O_NOCTTY
    TRUNC = lib.O_TRUNC
    APPEND = lib.O_APPEND
    NONBLOCK = lib.O_NONBLOCK
    DSYNC = lib.O_DSYNC
    DIRECT = lib.O_DIRECT
    LARGEFILE = lib.O_LARGEFILE
    DIRECTORY = lib.O_DIRECTORY
    NOFOLLOW = lib.O_NOFOLLOW
    NOATIME = lib.O_NOATIME
    CLOEXEC = lib.O_CLOEXEC
    SYNC = lib.O_SYNC
    PATH = lib.O_PATH
    TMPFILE = lib.O_TMPFILE
    # internal kernel flags, visible through FUSE and possibly other places
    FMODE_EXEC = 0x20
    FMODE_NONOTIFY = 0x4000000

class F(enum.IntEnum):
    "The cmd argument to fcntl; specifies what fcntl operation we want to do."
    SETFD = lib.F_SETFD
    GETFD = lib.F_GETFD
    SETFL = lib.F_SETFL
    ADD_SEALS = lib.F_ADD_SEALS
    GET_SEALS = lib.F_GET_SEALS

class F_SEAL(enum.IntEnum):
    "The single argument used with fcntl F.ADD_SEALS and F.GET_SEALS"
    SEAL = lib.F_SEAL_SEAL
    SHRINK = lib.F_SEAL_SHRINK
    GROW = lib.F_SEAL_GROW
    WRITE = lib.F_SEAL_WRITE
    FUTURE_WRITE = lib.F_SEAL_FUTURE_WRITE

class FD(enum.IntFlag):
    """The argument to fcntl F.SETFD and return value of fcntl F.GETFD

    This is totally different from `rsyscall.FileDescriptor`, despite
    "FD" being a frequently-used abbreviation for that class name.

    """
    CLOEXEC = lib.FD_CLOEXEC

#### Classes ####
from rsyscall.handle.fd import BaseFileDescriptor

class FcntlFileDescriptor(BaseFileDescriptor):
    async def fcntl(self, cmd: F, arg: t.Optional[int]=None) -> int:
        self._validate()
        try:
            return (await _fcntl(self.task.sysif, self.near, cmd, arg))
        except OSError as exn:
            exn.filename = (self, cmd, arg)
            raise

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _fcntl(sysif: SyscallInterface, fd: near.FileDescriptor,
                 cmd: F, arg: t.Optional[t.Union[int, near.Address]]=None) -> int:
    if arg is None:
        arg = 0
    return (await sysif.syscall(SYS.fcntl, fd, cmd, arg))



"""A thread-creating stub which can be exec'd by other programs which pass down stdin

The user creates an arbitrary Command which eventually execs rsyscall-stdin-bootstrap, and
calls stdin_bootstrap with it; stdin_bootstrap runs the command and returns a thread
created from the eventually-exec'd rsyscall-stdin-bootstrap.

This is useful for sudo and similar programs; we can pass a Command which runs "sudo
rsyscall-stdin-bootstrap", and thereby get a Thread with different privileges.

"""
import typing as t
from dataclasses import dataclass
from rsyscall.command import Command
from rsyscall.environ import Environment
from rsyscall.epoller import Epoller, AsyncReadBuffer
from rsyscall.handle import WrittenPointer, Task
from rsyscall.thread import Thread
from rsyscall.loader import NativeLoader
from rsyscall.memory.ram import RAM
from rsyscall.memory.socket_transport import SocketMemoryTransport
from rsyscall.monitor import AsyncChildProcess, ChildProcessMonitor
from rsyscall.tasks.connection import SyscallConnection
import logging
import rsyscall.far as far
import rsyscall.handle as handle
import rsyscall.memory.allocator as memory
import rsyscall.near.types as near
import rsyscall.nix as nix
import struct
import trio

from rsyscall.path import Path
from rsyscall.sched import CLONE
from rsyscall.signal import SIG, SignalBlock
from rsyscall.sys.mman import MFD
from rsyscall.sys.socket import SOCK, AF, SendmsgFlags, Socketpair, SendMsghdr, CmsgList, CmsgSCMRights
from rsyscall.sys.uio import IovecList

__all__ = [
    "stdin_bootstrap_path_with_nix",
    "stdin_bootstrap",
]

logger = logging.getLogger(__name__)

async def stdin_bootstrap_path_with_nix(thread: Thread) -> Path:
    """Get the path to the rsyscall-stdin-bootstrap executable.

    We return a Path rather than a Command because the typical usage
    of this path will be to pass it as an argument to some other
    command, such as sudo.

    """
    import rsyscall._nixdeps.librsyscall
    rsyscall_path = await nix.deploy(thread, rsyscall._nixdeps.librsyscall.closure)
    return rsyscall_path/"libexec"/"rsyscall"/"rsyscall-stdin-bootstrap"

async def stdin_bootstrap(
        parent: Thread,
        bootstrap_command: Command,
) -> t.Tuple[AsyncChildProcess, Thread]:
    """Create a thread from running an arbitrary command which must run rsyscall-stdin-bootstrap

    bootstrap_command can be any arbitrary command, but it must eventually exec
    rsyscall-stdin-bootstrap, and pass down stdin when it does.

    We'll clone and exec bootstrap_command, passing down a socketpair for stdin, and try to
    bootstrap over the other end of the socketpair. Once rsyscall-stdin-bootstrap starts,
    it will respond to our bootstrap and we'll create a new thread.

    """
    #### clone and exec into the bootstrap command
    # create the socketpair that will be used as stdin
    stdin_pair = await (await parent.task.socketpair(
        AF.UNIX, SOCK.STREAM, 0, await parent.ram.malloc(Socketpair))).read()
    parent_sock = stdin_pair.first
    child = await parent.clone()
    # set up stdin with socketpair
    await child.task.inherit_fd(stdin_pair.second).dup2(child.stdin)
    await stdin_pair.second.close()
    # exec
    child_process = await child.exec(bootstrap_command)
    #### set up all the fds we'll want to pass over
    # the basic connections
    [(access_syscall_sock, passed_syscall_sock),
     (access_data_sock, passed_data_sock)] = await parent.open_async_channels(2)
    # send the fds to the new process
    connection_fd, make_connection = await parent.connection.prep_fd_transfer()
    async def sendmsg_op(sem: RAM) -> WrittenPointer[SendMsghdr]:
        iovec = await sem.ptr(IovecList([await sem.malloc(bytes, 1)]))
        cmsgs = await sem.ptr(CmsgList([CmsgSCMRights([
            passed_syscall_sock, passed_data_sock, connection_fd])]))
        return await sem.ptr(SendMsghdr(None, iovec, cmsgs))
    _, [] = await parent_sock.sendmsg(await parent.ram.perform_batch(sendmsg_op), SendmsgFlags.NONE)
    # close our reference to fds that only the new process needs
    await passed_syscall_sock.close()
    await passed_data_sock.close()
    # close the socketpair
    await parent_sock.close()
    #### read describe to get all the information we need from the new process
    describe_buf = AsyncReadBuffer(access_data_sock)
    describe_struct = await describe_buf.read_cffi('struct rsyscall_stdin_bootstrap')
    environ = await describe_buf.read_envp(describe_struct.envp_count)
    #### build the new task
    pid = describe_struct.pid
    fd_table = handle.FDTable(pid)
    address_space = far.AddressSpace(pid)
    # we assume pid namespace is shared
    # TODO include namespace inode numbers numbers in describe
    # note: if we start dealing with namespace numbers then we need to
    # have a Kernel namespace which tells us which kernel we get those
    # numbers from.
    # oh hey we can conveniently dump the inode numbers with getdents!
    pidns = parent.task.pidns
    process = near.Process(pid)
    base_task = Task(process, fd_table, address_space, pidns)
    remote_syscall_fd = base_task.make_fd_handle(near.FileDescriptor(describe_struct.syscall_fd))
    base_task.sysif = SyscallConnection(
        logger.getChild(str(process)),
        access_syscall_sock, access_syscall_sock,
        remote_syscall_fd, remote_syscall_fd,
    )
    allocator = memory.AllocatorClient.make_allocator(base_task)
    # we assume our SignalMask is zero'd before being started, so we don't inherit it
    ram = RAM(base_task,
               SocketMemoryTransport(access_data_sock,
                                     base_task.make_fd_handle(near.FileDescriptor(describe_struct.data_fd))),
               allocator)
    # TODO I think I can maybe elide creating this epollcenter and instead inherit it or share it, maybe?
    epoller = await Epoller.make_root(ram, base_task)
    child_monitor = await ChildProcessMonitor.make(ram, base_task, epoller)
    connection = make_connection(base_task, ram,
                                 base_task.make_fd_handle(near.FileDescriptor(describe_struct.connecting_fd)))
    new_parent = Thread(
        task=base_task,
        ram=ram,
        connection=connection,
        loader=NativeLoader.make_from_symbols(base_task, describe_struct.symbols),
        epoller=epoller,
        child_monitor=child_monitor,
        environ=Environment.make_from_environ(base_task, ram, environ),
        stdin=base_task.make_fd_handle(near.FileDescriptor(0)),
        stdout=base_task.make_fd_handle(near.FileDescriptor(1)),
        stderr=base_task.make_fd_handle(near.FileDescriptor(2)),
    )
    return child_process, new_parent



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread, Command
from rsyscall.sys.wait import W
import typing as t
import cProfile
import pstats

async def rsys_run(cmd: Command, count: int) -> None:
    for _ in range(count):
        thread = await local_thread.clone()
        child = await thread.exec(cmd)
        await child.waitpid(W.EXITED)

class TestBenchmark(TrioTestCase):
    async def test_bench(self):
        "Run a simple clone and exec case inside cProfile (without asserting the results)"

        await local_thread.environ.as_arglist(local_thread.ram)
        cmd = (await local_thread.environ.which("echo")).args("-n")
        pr = cProfile.Profile()
        warm_up_runs = 1
        real_runs = 1
        await rsys_run(cmd, warm_up_runs)
        pr.enable()
        await rsys_run(cmd, real_runs)
        pr.disable()
        # pr.print_stats(sort='cumtime')
        # ps = pstats.Stats(pr).strip_dirs().sort_stats('cumulative')
        # ps.print_callees()



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
from rsyscall.signal import *

class TestSignal(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread

    async def test_sigaction(self) -> None:
        sa = Sigaction(Sighandler.DFL)
        ptr = await self.thr.ram.ptr(sa)
        await self.thr.task.sigaction(SIG.WINCH, ptr, None)
        await self.thr.task.sigaction(SIG.WINCH, None, ptr)
        out_sa = await ptr.read()
        self.assertEqual(sa.handler, out_sa.handler)
        self.assertEqual(sa.flags, out_sa.flags)
        self.assertEqual(sa.mask, out_sa.mask)
        self.assertEqual(sa.restorer, out_sa.restorer)

    # TODO test_signalblock
    # async def test_signalblock(self) -> None:
    #     pass



"""A thread on a remote host, bootstrapped over ssh

Note that all thread types can be launched through any other thread type,
including through an ssh thread.
"""
from __future__ import annotations
from dataclasses import dataclass
from rsyscall.command import Command
from rsyscall.environ import Environment
from rsyscall.epoller import Epoller, AsyncFileDescriptor, AsyncReadBuffer
from rsyscall.handle import WrittenPointer, FileDescriptor, Task
from rsyscall.thread import Thread
from rsyscall.loader import NativeLoader
from rsyscall.memory.ram import RAM
from rsyscall.memory.socket_transport import SocketMemoryTransport
from rsyscall.monitor import AsyncChildProcess, ChildProcessMonitor
from rsyscall.network.connection import ListeningConnection
from rsyscall.path import Path
from rsyscall.sched import CLONE
from rsyscall.tasks.connection import SyscallConnection
import abc
import contextlib
import importlib.resources
import logging
import os
import random
import rsyscall.far as far
import rsyscall.handle as handle
import rsyscall.memory.allocator as memory
import rsyscall.near.types as near
import rsyscall.nix as nix
import string
import typing as t

from rsyscall.fcntl import O, F
from rsyscall.stdlib import mkdtemp
from rsyscall.sys.socket import SOCK, AF
from rsyscall.sys.un import SockaddrUn
from rsyscall.sys.wait import W
from rsyscall.unistd import Pipe

__all__ = [
    "SSHCommand",
    "SSHDCommand",
    "SSHExecutables",
    "SSHDExecutables",
    "SSHHost",
    "make_local_ssh_from_executables",
    "make_ssh_host",
    "make_local_ssh",
]

ssh_bootstrap_script_contents = importlib.resources.read_text('rsyscall.tasks', 'ssh_bootstrap.sh')
logger = logging.getLogger(__name__)

T_ssh_command = t.TypeVar('T_ssh_command', bound="SSHCommand")
class SSHCommand(Command):
    "The 'ssh' executable provided by OpenSSH, plus some arguments and special methods"
    def ssh_options(self, config: t.Mapping[str, t.Union[str, bytes, os.PathLike]]) -> SSHCommand:
        option_list: t.List[str] = []
        for key, value in config.items():
            option_list += ["-o", os.fsdecode(key) + "=" + os.fsdecode(value)]
        return self.args(*option_list)

    def proxy_command(self, command: Command) -> SSHCommand:
        return self.ssh_options({'ProxyCommand': command.in_shell_form()})

    def local_forward(self, local_socket: str, remote_socket: str) -> SSHCommand:
        return self.args("-L", os.fsdecode(local_socket) + ":" + os.fsdecode(remote_socket))

    @classmethod
    def make(cls: t.Type[T_ssh_command], executable_path: Path) -> T_ssh_command:
        return cls(executable_path, ["ssh"], {})

class SSHDCommand(Command):
    "The 'sshd' executable provided by OpenSSH, plus some arguments and special methods"
    def sshd_options(self, config: t.Mapping[str, t.Union[str, bytes, os.PathLike]]) -> SSHDCommand:
        option_list: t.List[str] = []
        for key, value in config.items():
            option_list += ["-o", os.fsdecode(key) + "=" + os.fsdecode(value)]
        return self.args(*option_list)

    @classmethod
    def make(cls, executable_path: Path) -> SSHDCommand:
        return cls(executable_path, ["sshd"], {})

@dataclass
class SSHExecutables:
    """A standalone representation of the executables needed to create an SSH thread

    This is not really a user-facing class, it exists just to promote modularity. With
    this class, our functions need only take an object of this type, rather than look up
    the location of the executables themselves; therefore we can add new ways to look up
    executables and create this class without having to teach our functions about them.

    """
    base_ssh: SSHCommand
    bootstrap_path: Path

    @classmethod
    async def with_nix(cls, thread: Thread) -> SSHExecutables:
        import rsyscall._nixdeps.openssh
        import rsyscall._nixdeps.librsyscall
        ssh_path = await nix.deploy(thread, rsyscall._nixdeps.openssh.closure)
        rsyscall_path = await nix.deploy(thread, rsyscall._nixdeps.librsyscall.closure)
        base_ssh = SSHCommand.make(ssh_path/"bin"/"ssh")
        bootstrap_path = rsyscall_path/"libexec"/"rsyscall"/"rsyscall-bootstrap"
        return SSHExecutables(base_ssh, bootstrap_path)

    def host(self, to_host: t.Callable[[SSHCommand], SSHCommand]) -> SSHHost:
        """Create an object for sshing to a host.

        Important design decision here: the user doesn't pass in a
        hostname, username, various options, etc etc.

        Instead, they just give us a partial ssh command that we'll
        then use to do the sshing by appending our own shell command
        arguments.

        This allows using fancy options, connection sharing, all kinds
        of stuff, without us having to explicitly support it.

        The only constraint is that the user (obviously) shouldn't
        include an actual shell command in their ssh command.

        ---

        There's a further design decision here: We take a function
        instead of a completed SSHCommand. This is just so that we can
        pass in the basic SSHCommand to use, and then the user can
        extend it - it gives both us and the user the ability to add
        arbitrary exciting arguments.

        """
        return SSHHost(self, to_host)

class SSHHost:
    """A host we can ssh to, based on some ssh command

    We don't actually know what host we're going to ssh to - that's entirely determined by
    the user-provided to_host function. Presumably that function is deterministic, so
    we'll ssh to the same host each time...

    """
    def __init__(self,
                 executables: SSHExecutables,
                 to_host: t.Callable[[SSHCommand], SSHCommand]) -> None:
        self.executables = executables
        self.to_host = to_host

    async def ssh(self, thread: Thread) -> t.Tuple[AsyncChildProcess, Thread]:
        # we could get rid of the need to touch the local filesystem by directly
        # speaking the openssh multiplexer protocol. or directly speaking the ssh
        # protocol for that matter.
        ssh_to_host = self.to_host(self.executables.base_ssh)
        # we guess that the last argument of ssh command is the hostname. it
        # doesn't matter if it isn't, this is just used for a temp filename,
        # just to be more human-readable
        hostname = os.fsdecode(ssh_to_host.arguments[-1])
        random_suffix = ''.join(random.choices(string.ascii_uppercase + string.digits, k=8))
        name = (hostname+random_suffix+".sock")
        local_socket_path = thread.environ.tmpdir/name
        fd = await thread.task.open(await thread.ram.ptr(self.executables.bootstrap_path), O.RDONLY)
        async with make_bootstrap_dir(thread, ssh_to_host, fd) as tmp_path_bytes:
            return await ssh_bootstrap(thread, ssh_to_host, local_socket_path, tmp_path_bytes)

@contextlib.asynccontextmanager
async def make_bootstrap_dir(
        parent: Thread,
        ssh_command: SSHCommand,
        bootstrap_executable: FileDescriptor,
) -> t.AsyncGenerator[bytes, None]:
    """Over ssh, make a temporary directory containing the bootstrap executable, and start the socket bootstrap server

    The socket bootstrap server listens on two sockets in this temporary directory. One of
    them, we'll ssh forward back to the local host. The other, the main bootstrap process
    will connect to, to grab the listening socket fd for the former, so we can accept
    connections.

    We'll also use the bootstrap executable left in the temporary directory in
    ssh_bootstrap: we'll executed it to start the main bootstrap process.

    """
    stdout_pipe = await (await parent.task.pipe(
        await parent.ram.malloc(Pipe))).read()
    async_stdout = await parent.make_afd(stdout_pipe.read, set_nonblock=True)
    child = await parent.clone()
    await child.task.inherit_fd(stdout_pipe.write).dup2(child.stdout)
    await child.task.inherit_fd(bootstrap_executable).dup2(child.stdin)
    child_process = await child.exec(ssh_command.args(ssh_bootstrap_script_contents))
    await stdout_pipe.write.close()
    # from... local?
    # I guess this throws into sharper relief the distinction between core and module.
    # The ssh bootstrapping stuff should come from a different class,
    # which hardcodes the path,
    # and which works only for local tasks.
    # So in the meantime we'll continue to get it from task.filesystem.

    # sigh, openssh doesn't close its local stdout when it sees HUP/EOF on
    # the remote stdout. so we can't use EOF to signal end of our lines, and
    # instead have to have a sentinel to tell us when to stop reading.
    lines_buf = AsyncReadBuffer(async_stdout)
    tmp_path_bytes = await lines_buf.read_line()
    done = await lines_buf.read_line()
    if done != b"done":
        raise Exception("socket binder violated protocol, got instead of done:", done)
    await async_stdout.close()
    logger.debug("socket bootstrap done, got tmp path %s", tmp_path_bytes)
    yield tmp_path_bytes
    await child_process.check()

async def ssh_forward(thread: Thread, ssh_command: SSHCommand,
                      local_path: Path, remote_path: str) -> AsyncChildProcess:
    "Forward Unix socket connections to local_path to the socket at remote_path, over ssh"
    stdout_pipe = await (await thread.task.pipe(
        await thread.ram.malloc(Pipe))).read()
    async_stdout = await thread.make_afd(stdout_pipe.read, set_nonblock=True)
    child = await thread.clone()
    await child.task.inherit_fd(stdout_pipe.write).dup2(child.stdout)
    await child.task.chdir(await thread.ptr(local_path.parent))
    child_process = await child.exec(ssh_command.local_forward(
        "./" + local_path.name, remote_path,
    # TODO I optimistically assume that I'll have established a
    # connection through the tunnel before 1 minute has passed;
    # that connection will then keep the tunnel open.
    ).args("-n", "echo forwarded; exec sleep 60"))
    lines_buf = AsyncReadBuffer(async_stdout)
    forwarded = await lines_buf.read_line()
    if forwarded != b"forwarded":
        raise Exception("ssh forwarding violated protocol, got instead of forwarded:", forwarded)
    await async_stdout.close()
    return child_process

async def ssh_bootstrap(
        parent: Thread,
        # the actual ssh command to run
        ssh_command: SSHCommand,
        # the local path we'll use for the socket
        local_socket_path: Path,
        # the directory we're bootstrapping out of
        tmp_path_bytes: bytes,
) -> t.Tuple[AsyncChildProcess, Thread]:
    "Over ssh, run the bootstrap executable, "
    # identify local path
    local_data_addr = await parent.ram.ptr(
        await SockaddrUn.from_path(parent, local_socket_path))
    # start port forwarding; we'll just leak this process, no big deal
    # TODO we shouldn't leak processes; we should be GCing processes at some point
    forward_child_process = await ssh_forward(
        parent, ssh_command, local_socket_path, (tmp_path_bytes + b"/data").decode())
    # start bootstrap
    bootstrap_thread = await parent.clone()
    bootstrap_child_process = await bootstrap_thread.exec(ssh_command.args(
        "-n", f"cd {tmp_path_bytes.decode()}; exec ./bootstrap rsyscall"
    ))
    # TODO should unlink the bootstrap after I'm done execing.
    # it would be better if sh supported fexecve, then I could unlink it before I exec...
    # Connect to local socket 4 times
    async def make_async_connection() -> AsyncFileDescriptor:
        sock = await parent.make_afd(await parent.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
        await sock.connect(local_data_addr)
        return sock
    async_local_syscall_sock = await make_async_connection()
    async_local_data_sock = await make_async_connection()
    # Read description off of the data sock
    describe_buf = AsyncReadBuffer(async_local_data_sock)
    describe_struct = await describe_buf.read_cffi('struct rsyscall_bootstrap')
    new_pid = describe_struct.pid
    environ = await describe_buf.read_envp(describe_struct.envp_count)
    # Build the new task!
    new_address_space = far.AddressSpace(new_pid)
    # TODO the pid namespace will probably be common for all connections...
    # TODO we should get this from the SSHHost, this is usually going
    # to be common for all connections and we should express that
    new_pid_namespace = far.PidNamespace(new_pid)
    new_process = near.Process(new_pid)
    new_base_task = Task(
        new_process, handle.FDTable(new_pid), new_address_space,
        new_pid_namespace,
    )
    handle_remote_syscall_fd = new_base_task.make_fd_handle(near.FileDescriptor(describe_struct.syscall_sock))
    new_base_task.sysif = SyscallConnection(
        logger.getChild(str(new_process)),
        async_local_syscall_sock, async_local_syscall_sock,
        handle_remote_syscall_fd, handle_remote_syscall_fd,
    )
    handle_remote_data_fd = new_base_task.make_fd_handle(near.FileDescriptor(describe_struct.data_sock))
    handle_listening_fd = new_base_task.make_fd_handle(near.FileDescriptor(describe_struct.listening_sock))
    new_allocator = memory.AllocatorClient.make_allocator(new_base_task)
    new_transport = SocketMemoryTransport(async_local_data_sock, handle_remote_data_fd)
    # we don't inherit SignalMask; we assume ssh zeroes the sigmask before starting us
    new_ram = RAM(new_base_task, new_transport, new_allocator)
    epoller = await Epoller.make_root(new_ram, new_base_task)
    child_monitor = await ChildProcessMonitor.make(new_ram, new_base_task, epoller)
    await handle_listening_fd.fcntl(F.SETFL, O.NONBLOCK)
    connection = ListeningConnection(
        parent.task, parent.ram, parent.epoller,
        local_data_addr,
        new_base_task, new_ram,
        await AsyncFileDescriptor.make(epoller, new_ram, handle_listening_fd),
    )
    new_thread = Thread(
        task=new_base_task,
        ram=new_ram,
        connection=connection,
        loader=NativeLoader.make_from_symbols(new_base_task, describe_struct.symbols),
        epoller=epoller,
        child_monitor=child_monitor,
        environ=Environment.make_from_environ(new_base_task, new_ram, environ),
        stdin=new_base_task.make_fd_handle(near.FileDescriptor(0)),
        stdout=new_base_task.make_fd_handle(near.FileDescriptor(1)),
        stderr=new_base_task.make_fd_handle(near.FileDescriptor(2)),
    )
    return bootstrap_child_process, new_thread

@dataclass
class SSHDExecutables:
    """A standalone representation of the executables needed to run sshd

    This is not really a user-facing class; see SSHExecutables.

    """
    ssh_keygen: Command
    sshd: SSHDCommand

    @classmethod
    async def with_nix(cls, thread: Thread) -> SSHDExecutables:
        import rsyscall._nixdeps.openssh
        ssh_path = await nix.deploy(thread, rsyscall._nixdeps.openssh.closure)
        ssh_keygen = ssh_path.bin('ssh-keygen')
        sshd = SSHDCommand.make(ssh_path/"bin"/"sshd")
        return SSHDExecutables(ssh_keygen, sshd)

async def make_local_ssh_from_executables(thread: Thread,
                                          executables: SSHExecutables, sshd_executables: SSHDExecutables) -> SSHHost:
    "Make an SSHHost which just sshs to localhost; useful for testing"
    ssh_keygen = sshd_executables.ssh_keygen
    sshd = sshd_executables.sshd

    keygen_command = ssh_keygen.args('-b', '1024', '-q', '-N', '', '-C', '', '-f', 'key')
    keygen_thread = await thread.clone()
    # ugh, we have to make a directory because ssh-keygen really wants to output to a directory
    async with (await mkdtemp(thread)) as tmpdir:
        await keygen_thread.task.chdir(await keygen_thread.ram.ptr(tmpdir))
        await (await keygen_thread.exec(keygen_command)).waitpid(W.EXITED)
        privkey_file = await thread.task.open(await thread.ram.ptr(tmpdir/'key'), O.RDONLY)
        pubkey_file = await thread.task.open(await thread.ram.ptr(tmpdir/'key.pub'), O.RDONLY)
    def to_host(ssh: SSHCommand, privkey_file=privkey_file, pubkey_file=pubkey_file) -> SSHCommand:
        privkey = privkey_file.as_proc_path()
        pubkey = pubkey_file.as_proc_path()
        sshd_command = sshd.args(
            '-i', '-e', '-f', '/dev/null',
        ).sshd_options({
            'LogLevel': 'INFO',
            'HostKey': privkey,
            'AuthorizedKeysFile': pubkey,
            'StrictModes': 'no',
            'PrintLastLog': 'no',
            'PrintMotd': 'no',
            'UsePrivilegeSeparation': 'no',
        })
        ssh_command = ssh.args(
            '-F', '/dev/null',
        ).ssh_options({
            'LogLevel': 'INFO',
            'IdentityFile': privkey,
            'BatchMode': 'yes',
            'StrictHostKeyChecking': 'no',
            'UserKnownHostsFile': '/dev/null',
        }).proxy_command(sshd_command).args(
            "localhost",
        )
        return ssh_command
    return executables.host(to_host)

# Helpers
async def make_ssh_host(thread: Thread, to_host: t.Callable[[SSHCommand], SSHCommand]) -> SSHHost:
    ssh = await SSHExecutables.with_nix(thread)
    return ssh.host(to_host)

async def make_local_ssh(thread: Thread) -> SSHHost:
    "Look up the ssh executables and return an SSHHost which sshs to localhost; useful for testing"
    ssh = await SSHExecutables.with_nix(thread)
    sshd = await SSHDExecutables.with_nix(thread)
    return (await make_local_ssh_from_executables(thread, ssh, sshd))




import typing as t
import trio
from rsyscall.epoller import Epoller, AsyncFileDescriptor
from rsyscall.unistd import Pipe
from rsyscall.fcntl import O
from rsyscall.thread import Thread

import logging
logger = logging.getLogger(__name__)
# logging.basicConfig(level=logging.DEBUG)

import unittest
async def do_async_things(self: unittest.TestCase, epoller: Epoller, thr: Thread, i: int=0,
                          *, task_status=trio.TASK_STATUS_IGNORED) -> None:
    logger.info("Setting up for do_async_things(%d)", i)
    pipe = await (await thr.task.pipe(await thr.ram.malloc(Pipe), O.NONBLOCK)).read()
    async_pipe_rfd = await AsyncFileDescriptor.make(epoller, thr.ram, pipe.read)
    async_pipe_wfd = await AsyncFileDescriptor.make(epoller, thr.ram, pipe.write)
    task_status.started(None)
    data = b"hello world"
    logger.info("Starting do_async_things(%d)", i)
    async def stuff():
        logger.info("do_async_things(%d): read(%s): starting", i, async_pipe_rfd.handle.near)
        result = await async_pipe_rfd.read_some_bytes()
        logger.info("do_async_things(%d): read(%s): returned", i, async_pipe_rfd.handle.near)
        self.assertEqual(result, data)
    async with trio.open_nursery() as nursery:
        nursery.start_soon(stuff)
        await trio.sleep(0.0001)
        # hmmm MMM MMMmmmm MMM mmm MMm mm MM mmm MM mm MM
        # does this make sense?
        logger.info("do_async_things(%d): write(%s): starting", i, async_pipe_wfd.handle.near)
        await async_pipe_wfd.write_all_bytes(data)
        logger.info("do_async_things(%d): write(%s): returned", i, async_pipe_wfd.handle.near)
    await async_pipe_rfd.close()
    await async_pipe_wfd.close()
    logger.info("Done with do_async_things(%d)", i)

async def assert_thread_works(self: unittest.TestCase, thr: Thread) -> None:
    await do_async_things(self, thr.epoller, thr)



"`#include <netinet/tcp.h>`"
from rsyscall._raw import ffi, lib # type: ignore
import enum

__all__ = [
    'TCP',
]

class TCP(enum.IntEnum):
    "User-settable options (used with setsockopt)"
    NODELAY = lib.TCP_NODELAY                           # Don't delay send to coalesce packets
    MAXSEG = lib.TCP_MAXSEG                             # Set maximum segment size
    CORK = lib.TCP_CORK                                 # Control sending of partial frames
    KEEPIDLE = lib.TCP_KEEPIDLE                         # Start keeplives after this period
    KEEPINTVL = lib.TCP_KEEPINTVL                       # Interval between keepalives
    KEEPCNT = lib.TCP_KEEPCNT                           # Number of keepalives before death
    SYNCNT = lib.TCP_SYNCNT                             # Number of SYN retransmits
    LINGER2 = lib.TCP_LINGER2                           # Life time of orphaned FIN-WAIT-2 state
    DEFER_ACCEPT = lib.TCP_DEFER_ACCEPT                 # Wake up listener only when data arrive
    WINDOW_CLAMP = lib.TCP_WINDOW_CLAMP                 # Bound advertised window
    INFO = lib.TCP_INFO                                 # Information about this connection.
    QUICKACK = lib.TCP_QUICKACK                         # Bock/reenable quick ACKs.
    CONGESTION = lib.TCP_CONGESTION                     # Congestion control algorithm.
    MD5SIG = lib.TCP_MD5SIG                             # TCP MD5 Signature (RFC2385)
    COOKIE_TRANSACTIONS = lib.TCP_COOKIE_TRANSACTIONS   # TCP Cookie Transactions
    THIN_LINEAR_TIMEOUTS = lib.TCP_THIN_LINEAR_TIMEOUTS # Use linear timeouts for thin streams
    THIN_DUPACK = lib.TCP_THIN_DUPACK                   # Fast retrans. after 1 dupack
    USER_TIMEOUT = lib.TCP_USER_TIMEOUT                 # How long for loss retry before timeout
    REPAIR = lib.TCP_REPAIR                             # TCP sock is under repair right now
    REPAIR_QUEUE = lib.TCP_REPAIR_QUEUE                 # Set TCP queue to repair
    QUEUE_SEQ = lib.TCP_QUEUE_SEQ                       # Set sequence number of repaired queue.
    REPAIR_OPTIONS = lib.TCP_REPAIR_OPTIONS             # Repair TCP connection options
    FASTOPEN = lib.TCP_FASTOPEN                         # Enable FastOpen on listeners
    TIMESTAMP = lib.TCP_TIMESTAMP                       # TCP time stamp
    NOTSENT_LOWAT = lib.TCP_NOTSENT_LOWAT               # Limit number of unsent bytes in write queue.
    CC_INFO = lib.TCP_CC_INFO                           # Get Congestion Control (optional) info.
    SAVE_SYN = lib.TCP_SAVE_SYN                         # Record SYN headers for new connections.
    SAVED_SYN = lib.TCP_SAVED_SYN                       # Get SYN headers recorded for connection.
    REPAIR_WINDOW = lib.TCP_REPAIR_WINDOW               # Get/set window parameters.
    FASTOPEN_CONNECT = lib.TCP_FASTOPEN_CONNECT         # Attempt FastOpen with connect.
    ULP = lib.TCP_ULP                                   # Attach a ULP to a TCP connection.
    MD5SIG_EXT = lib.TCP_MD5SIG_EXT                     # TCP MD5 Signature with extensions.
    FASTOPEN_KEY = lib.TCP_FASTOPEN_KEY                 # Set the key for Fast Open (cookie).
    FASTOPEN_NO_COOKIE = lib.TCP_FASTOPEN_NO_COOKIE     # Enable TFO without a TFO cookie.
    ZEROCOPY_RECEIVE = lib.TCP_ZEROCOPY_RECEIVE         # Perform a zerocopy receive
    INQ = lib.TCP_INQ                                   # Notify bytes available to read as a cmsg on read.
    CM_INQ = lib.TCP_CM_INQ                             # CMSG type for bytes available
    TX_DELAY = lib.TCP_TX_DELAY                         # Delay outgoing packets by XX usec.



"""Persistent threads, which live on after their parent thread has died, and to which we can reconnect

These are threads which stick around, holding resources, even if we died or our connection
to them fails. We can reconnect to the persistent thread and continue to access all the
old resources.

After we disconnect from a persistent thread for whatever reason, it is left in an inert
state. It won't make any syscalls until we reconnect to it.

This is useful for fault isolation. To run and monitor long-running processes, we can and
should ourselves be long-running; but, we might crash at some point. The persistent thread
provides fault isolation: We may crash, but the thread and all its children will stay
alive, and we can reconnect to the thread after restarting and resume monitoring the
long-running processes. The persistent thread provides similar fault isolation for any
other resource that a thread might hold, such as file descriptors; the persistent thread
will keep file descriptors open even after we die.

The persistent thread doesn't provide any mechanism to recover information about the state
of the persistent thread, such as what child processes it has and what file descriptors it
has open. Therefore, we should arrange to have other state-persistence mechanism so that
we can recover our state if we crash. For example, we could maintain an on-disk database
where we log information about what processes we've started in what persistent threads,
and what the status of each of those processes is. Then we could recover information about
the state of the persistent thread from that database, reconnect to the persistent thread,
and resume normal operation.

That being said, a caveat: We don't currently have any logic to support recovering
information from such a database after a crash in a safe way. Such a database could be
created, but it would require creating resource handles in a way that is not guaranteed to
be safe. Determining a clean, generic way to persist information about the state of
resources in a way that can be safely recovered after a crash is an open question.

## `CLONE_THREAD`

The model we currently use for persistent threads is:

1. Create this can-be-persistent thread
2. Do a bunch of things in that thread, allocating whatever resources
3. Call make_persistent to make the thread actually persistent
4. Crash or disconnect, and call reconnect to reconnect.

It would be better for the model to be:

1. Do a bunch of things in whatever thread you like, allocating whatever resources
2. Create an immediately persistent thread which inherits those resources
3. Crash or disconnect, and call reconnect to reconnect.

However, the major obstacle is child processes. Child processes can't be inherited to a
new child process, much less passed around between unrelated processes like file
descriptors can.

`CLONE_THREAD` allows creating a new child process which can wait on the child processes of the
parent; however, `CLONE_THREAD` also does a bunch of other stuff which is undesirable. Among
other things, `CLONE_THREAD` processes:
- don't send `SIGCHLD` when exiting, so they can't be waited on without dedicating a thread
  to monitor them
- don't leave a zombie when they die
- block several unshare and setns operations
- complicate signals and many other system calls

While `CLONE_THREAD` could allow the better model for persistent threads, it comes with a
host of other disadvantages and complexities, so we're just biting the bullet and
accepting the worse model.

The recent feature of pidfds will likely allow a better model, once we're able to wait on
pidfds, which hasn't yet been added.

"""
from __future__ import annotations
from dneio import Continuation, shift
import typing as t
import rsyscall.near.types as near
import rsyscall.far as far
import rsyscall.handle as handle
from rsyscall.thread import Thread
from rsyscall.tasks.connection import SyscallConnection
from rsyscall.tasks.clone import clone_child_task
from rsyscall.loader import NativeLoader, Trampoline
from rsyscall.sched import Stack
from rsyscall.handle import WrittenPointer, ThreadProcess, Pointer, Task, FileDescriptor
from rsyscall.memory.socket_transport import SocketMemoryTransport
from rsyscall.near.sysif import SyscallInterface, SyscallSendError
from rsyscall.sys.syscall import SYS

import trio
import struct
import os
from dataclasses import dataclass
import logging
from rsyscall.memory.ram import RAM

from rsyscall.monitor import ChildProcessMonitor
from rsyscall.epoller import Epoller, AsyncFileDescriptor
from rsyscall.sys.epoll import EPOLL

from rsyscall.struct import Int32, StructList

from rsyscall.sched import CLONE
from rsyscall.sys.socket import AF, SOCK, SendmsgFlags, SendMsghdr, CmsgSCMRights, CmsgList, SHUT
from rsyscall.sys.un import SockaddrUn
from rsyscall.sys.uio import IovecList
from rsyscall.signal import SIG, Sigset, SignalBlock
from rsyscall.sys.prctl import PR

__all__ = [
    "clone_persistent",
    "PersistentThread",
]

logger = logging.getLogger(__name__)

class PersistentSyscallConnection(SyscallInterface):
    def __init__(self, conn: SyscallConnection) -> None:
        self.conn: t.Optional[SyscallConnection] = conn
        self._activity_fd_conn: t.Optional[SyscallConnection] = conn
        self.logger = self.conn.logger
        self._new_conn_cbs: t.List[Continuation[SyscallConnection]] = []
        self._break_exc: t.Optional[SyscallSendError] = None
        self._break_cb: t.Optional[Continuation[SyscallSendError]] = None

    def wait_for_break_cb(self, cb: Continuation[SyscallSendError]) -> None:
        self._break_cb = cb

    async def wait_for_break(self) -> SyscallSendError:
        assert self._break_cb is None
        if self._break_exc:
            exc = self._break_exc
            self._break_exc = None
            return exc
        else:
            return await shift(self.wait_for_break_cb)

    def set_new_conn(self, conn: SyscallConnection) -> None:
        # to preserve the global relative ordering of syscalls,
        # we don't set self.conn again until syscalls stop appearing on self._new_conn_cbs
        # but we have to set activity_fd_conn because some callback might use get_activity_fd
        self._activity_fd_conn = conn
        while self._new_conn_cbs:
            cbs = self._new_conn_cbs
            self._new_conn_cbs = []
            for cb in cbs:
                cb.send(conn)
        self.conn = conn

    async def shutdown_current_connection(self) -> None:
        # Shut down write end of the current connection; any currently running or pending
        # syscalls (such as epoll_wait) will be able to finish and receive their response,
        # after which the syscall server will close its end of the current connection and
        # start listening for a new connection.
        if self.conn:
            await self.conn.tofd.handle.shutdown(SHUT.WR)

    async def close_interface(self) -> None:
        if self.conn:
            await self.conn.close_interface()

    def get_activity_fd(self) -> FileDescriptor:
        if self._activity_fd_conn:
            return self._activity_fd_conn.get_activity_fd()
        raise Exception("can't get activity fd while disconnected")

    async def syscall(self, number: SYS, arg1=0, arg2=0, arg3=0, arg4=0, arg5=0, arg6=0) -> int:
        while True:
            if self.conn is None:
                conn = await shift(self._new_conn_cbs.append)
            else:
                conn = self.conn
            try:
                return await conn.syscall(number, arg1, arg2, arg3, arg4, arg5, arg6)
            except SyscallSendError as exc:
                if self.conn is conn:
                    self.conn = None
                    self._activity_fd_conn = None
                    if self._break_cb:
                        self._break_cb.send(exc)
                    else:
                        self._break_exc = exc

# this should be a method, I guess, on something which points to the persistent stuff resource.
async def clone_persistent(
        parent: Thread, path: t.Union[str, os.PathLike],
) -> PersistentThread:
    """Create a new not-yet-persistent thread and return the thread and its tracking object

    To make the thread actually persistent, you must call PersistentServer.make_persistent().

    The point of this hoop-jumping is just to prevent unnecessary resource leakage, so you
    can set up things in a persistent thread and only make it persistent when you're
    actually ready.

    A persistent thread is essentially the same as a normal thread, just running a
    different function. As such, it starts off sharing its file descriptor table and
    everything else with its parent thread. It's only when we disconnect and reconnect
    that it changes behavior.

    """
    listening_sock = await parent.task.socket(AF.UNIX, SOCK.STREAM)
    await listening_sock.bind(await parent.ram.ptr(await SockaddrUn.from_path(parent, path)))
    await listening_sock.listen(1)
    child_process, task = await clone_child_task(
        parent.task, parent.ram, parent.connection, parent.loader, parent.monitor,
        CLONE.FILES|CLONE.FS|CLONE.SIGHAND,
        lambda sock: Trampoline(parent.loader.persistent_server_func, [sock, sock, listening_sock]))
    listening_sock_handle = listening_sock.move(task)
    ram = RAM(task, parent.ram.transport, parent.ram.allocator.inherit(task))

    ## create the new persistent task
    epoller = await Epoller.make_root(ram, task)
    signal_block = SignalBlock(task, await ram.ptr(Sigset({SIG.CHLD})))
    # TODO use an inherited signalfd instead
    child_monitor = await ChildProcessMonitor.make(ram, task, epoller, signal_block=signal_block)
    return PersistentThread(Thread(
        task, ram,
        parent.connection.inherit(task, ram),
        parent.loader,
        epoller,
        child_monitor,
        parent.environ.inherit(task, ram),
        stdin=parent.stdin.for_task(task),
        stdout=parent.stdout.for_task(task),
        stderr=parent.stderr.for_task(task),
    ), persistent_path=path, persistent_sock=listening_sock_handle)

async def _connect_and_send(self: PersistentThread, thread: Thread, fds: t.List[FileDescriptor]) -> t.List[FileDescriptor]:
    """Connect to a persistent thread's socket, send some file descriptors

    This isn't actually a generic function; the persistent thread expects exactly three
    file descriptors, and uses them in a special way.

    """
    sock = await thread.make_afd(await thread.socket(AF.UNIX, SOCK.STREAM|SOCK.NONBLOCK))
    sockaddr_un = await SockaddrUn.from_path(thread, self.persistent_path)
    async def sendmsg_op(sem: RAM) -> t.Tuple[
            WrittenPointer[SockaddrUn], WrittenPointer[Int32], WrittenPointer[SendMsghdr], Pointer[StructList[Int32]]]:
        addr = await sem.ptr(sockaddr_un)
        count = await sem.ptr(Int32(len(fds)))
        iovec = await sem.ptr(IovecList([await sem.malloc(bytes, 1)]))
        cmsgs = await sem.ptr(CmsgList([CmsgSCMRights(fds)]))
        hdr = await sem.ptr(SendMsghdr(None, iovec, cmsgs))
        response_buf = await sem.ptr(StructList(Int32, [Int32(0)]*len(fds)))
        return addr, count, hdr, response_buf
    addr, count, hdr, response = await thread.ram.perform_batch(sendmsg_op)
    data = None
    await sock.connect(addr)
    _, _ = await sock.write(count)
    _, [] = await sock.handle.sendmsg(hdr, SendmsgFlags.NONE)
    while response.size() > 0:
        valid, response = await sock.read(response)
        data += valid
    remote_fds = [self.task.make_fd_handle(near.FileDescriptor(int(i)))
                  for i in ((await data.read()).elems if data else [])]
    await sock.close()
    return remote_fds

class PersistentThread(Thread):
    """A thread which can live on even if everything else has exited

    It's not persistent by default - you need to call make_persistent() first to make that
    happen. After that, this thread will continue living even if its parent dies or our
    connection to it fails, and you can reconnect to it by calling reconnect(thr), passing
    the thread you want to initiate the connection from.

    A great name for this would be "daemon thread", but that's already taken by more
    conventional thread systems to refer to a much more conventional kind of thread.  I
    wistfully recall the name I gave to a previous attempt at making a hosting system for
    long-running tasks: daemon engines. That was a great name.

    """
    def __init__(self,
                 thread: Thread,
                 persistent_path: t.Union[str, os.PathLike],
                 persistent_sock: FileDescriptor,
    ) -> None:
        super()._init_from(thread)
        self.persistent_path = persistent_path
        self.persistent_sock = persistent_sock
        self.prepped_for_reconnect = False

    async def prep_for_reconnect(self) -> None:
        # TODO hmm should we switch the transport?
        # i guess we aren't actually doing anything but rearranging the file descriptors
        await self.unshare_files(going_to_exec=False)
        if not isinstance(self.task.sysif, SyscallConnection):
            raise Exception("self.task.sysif of unexpected type", self.task.sysif)
        self.task.sysif = PersistentSyscallConnection(self.task.sysif)
        self.prepped_for_reconnect = True

    async def make_persistent(self) -> None:
        "Make this thread actually persistent"
        await self.prep_for_reconnect()
        await self.task.setsid()
        await self.task.prctl(PR.SET_PDEATHSIG, 0)

    async def reconnect(self, thread: Thread) -> None:
        """Using the passed-in thread to establish the connection, reconnect to this PersistentThread

        """
        if not self.prepped_for_reconnect:
            # It does work to reconnect without prep_for_reconnect, except for one nitpick:
            # If the underlying process for the PersistentThread dies while we're in the
            # middle of reconnecting to it, the file descriptors opened by the C code
            # running in the process will leak if the process is in a shared fd table.
            # That's annoying on its own, but also means we won't get an EOF from our
            # communication with the process, and we'll just hang forever.
            await self.prep_for_reconnect()
        await self.task.run_fd_table_gc(use_self=False)
        if not isinstance(self.task.sysif, PersistentSyscallConnection):
            raise Exception("self.task.sysif of unexpected type", self.task.sysif)
        await self.task.sysif.shutdown_current_connection()
        [(access_syscall_sock, syscall_sock), (access_data_sock, data_sock)] = await thread.open_async_channels(2)
        [infd, outfd, remote_data_sock] = await _connect_and_send(self, thread, [syscall_sock, syscall_sock, data_sock])
        await syscall_sock.close()
        await data_sock.close()
        # Set up the new SyscallConnection
        conn = SyscallConnection(
            self.task.sysif.logger,
            access_syscall_sock, access_syscall_sock,
            infd, outfd,
        )
        self.task.sysif.set_new_conn(conn)
        # Fix up RAM with new transport
        # TODO technically this could still be in the same address space - that's the case in our tests.
        # we should figure out a way to use a LocalMemoryTransport here so it can copy efficiently
        transport = SocketMemoryTransport(access_data_sock, remote_data_sock)
        self.ram.transport = transport
        self.transport = transport
        # close remote fds we don't have handles to; this includes the old interface fds.
        await self.task.run_fd_table_gc()



"Named for the excellent manpage about these process attributes, credentials(7)"
from __future__ import annotations
import rsyscall.near as near
import typing as t

#### Classes ####
if t.TYPE_CHECKING:
    from rsyscall.handle.process import ChildProcess
import rsyscall.far

class CredentialsTask(rsyscall.far.Task):
    async def getuid(self) -> int:
        return (await _getuid(self.sysif))

    async def getgid(self) -> int:
        return (await _getgid(self.sysif))

    async def getpid(self) -> near.Process:
        return (await _getpid(self.sysif))

    async def getpgid(self) -> near.ProcessGroup:
        return (await _getpgid(self.sysif, None))

    async def setpgid(self, pgid: t.Optional[ChildProcess]=None) -> None:
        if pgid is None:
            await _setpgid(self.sysif, None, None)
        else:
            if pgid.task.pidns != self.pidns:
                raise rsyscall.far.NamespaceMismatchError(
                    "different pid namespaces", pgid.task.pidns, self.pidns)
            with pgid.borrow():
                await _setpgid(self.sysif, None, pgid._as_process_group())

    async def setsid(self) -> int:
        return (await _setsid(self.sysif))

#### Raw syscalls ####
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _getuid(sysif: SyscallInterface) -> int:
    return (await sysif.syscall(SYS.getuid))

async def _getgid(sysif: SyscallInterface) -> int:
    return (await sysif.syscall(SYS.getgid))

async def _getpid(sysif: SyscallInterface) -> near.Process:
    return near.Process(await sysif.syscall(SYS.getpid))

async def _getpgid(sysif: SyscallInterface, pid: t.Optional[near.Process]) -> near.ProcessGroup:
    if pid is None:
        pid = 0 # type: ignore
    return near.ProcessGroup(await sysif.syscall(SYS.getpgid, pid))

async def _setpgid(sysif: SyscallInterface,
                   pid: t.Optional[near.Process], pgid: t.Optional[near.ProcessGroup]) -> None:
    if pid is None:
        pid = 0 # type: ignore
    if pgid is None:
        pgid = 0 # type: ignore
    await sysif.syscall(SYS.setpgid, pid, pgid)

async def _setsid(sysif: SyscallInterface) -> int:
    return (await sysif.syscall(SYS.setsid))



"""The local thread, based on the local Python interpreter thread.

The local thread is the thread that every rsyscall program has available from the
start. From this thread, we create all the others.

"""
from __future__ import annotations
from dneio.core import TrioSystemWaitReadable, set_trio_system_wait_readable
from rsyscall.thread import Thread
from rsyscall._raw import ffi, lib # type: ignore
import trio
import rsyscall.far as far
from rsyscall.near.sysif import SyscallInterface, Syscall, raise_if_error
import rsyscall.near.types as near
import rsyscall.handle as handle
import rsyscall.loader as loader
import logging
import os
import typing as t
from dataclasses import dataclass
import rsyscall.memory.allocator as memory
from rsyscall.memory.ram import RAM
from rsyscall.memory.transport import MemoryTransport
from rsyscall.handle import Pointer, Task
from rsyscall.signal import SIG, Sigaction, Sighandler
from rsyscall.sys.socket import AF, SOCK
from rsyscall.sys.syscall import SYS
from rsyscall.network.connection import FDPassConnection
from rsyscall.environ import Environment
from rsyscall.loader import NativeLoader
from rsyscall.monitor import ChildProcessMonitor

from rsyscall.sys.epoll import EpollFlag
from rsyscall.epoller import Epoller

logger = logging.getLogger(__name__)

async def _direct_syscall(number, arg1=0, arg2=0, arg3=0, arg4=0, arg5=0, arg6=0):
    "Make a syscall directly in the current thread."
    return lib.rsyscall_raw_syscall(arg1, arg2, arg3, arg4, arg5, arg6, number)

class LocalSyscall(SyscallInterface):
    "Makes syscalls in the local, Python interpreter thread."
    def __init__(self) -> None:
        self.logger = logger

    def get_activity_fd(self) -> None:
        return None

    async def close_interface(self) -> None:
        pass

    async def syscall(self, number: SYS, arg1=0, arg2=0, arg3=0, arg4=0, arg5=0, arg6=0) -> int:
        syscall = Syscall(number, arg1, arg2, arg3, arg4, arg5, arg6)
        self.logger.debug("%s", syscall)
        result = await _direct_syscall(
            number,
            arg1=int(arg1), arg2=int(arg2), arg3=int(arg3),
            arg4=int(arg4), arg5=int(arg5), arg6=int(arg6))
        try:
            raise_if_error(result)
        except OSError as exn:
            self.logger.debug("%s -> %s", number, exn)
            raise
        self.logger.debug("%s -> %s", number, result)
        return result

class LocalMemoryTransport(MemoryTransport):
    "This is a memory transport that only works on local pointers."
    def __init__(self, local_task: Task) -> None:
        self.local_task = local_task

    def inherit(self, task: Task) -> LocalMemoryTransport:
        return self

    async def write(self, dest: Pointer, data: bytes) -> None:
        if dest.mapping.task.address_space != self.local_task.address_space:
            raise Exception("trying to write to pointer", dest, "not in local address space")
        ffi.memmove(ffi.cast('void*', int(dest.near)), data, len(data))

    async def read(self, src: Pointer) -> bytes:
        if src.mapping.task.address_space != self.local_task.address_space:
            raise Exception("trying to read from pointer", src, "not in local address space")
        buf = ffi.buffer(ffi.cast('void*', int(src.near)), src.size())
        return bytes(buf)

async def _make_local_thread() -> Thread:
    """Create the local thread, allocating various resources locally.

    For the most part, the local thread is like any other thread; it just bootstraps
    differently, and uses syscall and memory interfaces which are specialized to the local
    thread.

    """
    process = near.Process(os.getpid())
    task = Task(
        process, handle.FDTable(process.id),
        far.AddressSpace(process.id),
        far.PidNamespace(process.id),
    )
    task.sysif = LocalSyscall()
    ram = RAM(task, LocalMemoryTransport(task), memory.AllocatorClient.make_allocator(task))
    epfd = await task.epoll_create()
    async def wait_readable():
        logger.debug("wait_readable(%s)", epfd.near.number)
        await trio.lowlevel.wait_readable(epfd.near.number)
    trio_system_wait_readable = TrioSystemWaitReadable(epfd.near.number)
    set_trio_system_wait_readable(trio_system_wait_readable)
    epoller = Epoller.make_subsidiary(ram, epfd, trio_system_wait_readable.wait)
    thread = Thread(
        task, ram,
        await FDPassConnection.make(task, ram, epoller),
        NativeLoader.make_from_symbols(task, lib),
        epoller,
        await ChildProcessMonitor.make(ram, task, epoller),
        Environment.make_from_environ(task, ram, {**os.environ}),
        stdin=task.make_fd_handle(near.FileDescriptor(0)),
        stdout=task.make_fd_handle(near.FileDescriptor(1)),
        stderr=task.make_fd_handle(near.FileDescriptor(2)),
    )
    return thread

async def _initialize() -> Thread:
    thr = await _make_local_thread()
    # Wipe out the SIGWINCH handler that the readline module installs.
    # We do this because otherwise this handler will be inherited down to our
    # children, where it will segfault on run due to the environment being
    # totally different (lacking TLS for one). I'm not sure if there's any
    # alternative; wiping out the signal handlers from within the children after
    # they've been created still leaves a window for the signal handler to run.
    import readline
    await thr.task.sigaction(SIG.WINCH, await thr.ram.ptr(Sigaction(Sighandler.DFL)), None)
    return thr

local_thread: Thread = trio.run(_initialize)
"The local thread, fully initialized at import time"



"`#include <netinet/ip.h>`"
from rsyscall._raw import ffi, lib # type: ignore
import enum

# ip.h is a superset of in.h
from rsyscall.netinet.in_ import SockaddrIn, SockaddrIn6
__all__ = [
    "SockaddrIn",
    "SockaddrIn6",
    "IPPROTO",
    "IP",
]

class IPPROTO(enum.IntEnum):
    "Used for a variety of things"
    ICMPV6 = lib.IPPROTO_ICMPV6

class IP(enum.IntEnum):
    "Mostly for socket options"
    RECVERR = lib.IP_RECVERR
    PKTINFO = lib.IP_PKTINFO
    MULTICAST_TTL = lib.IP_MULTICAST_TTL
    MTU_DISCOVER = lib.IP_MTU_DISCOVER
    PMTUDISC_DONT = lib.IP_PMTUDISC_DONT



from __future__ import annotations
from rsyscall._raw import ffi # type: ignore
from dataclasses import dataclass
from rsyscall.command import Command
from rsyscall.handle.pointer import Pointer, WrittenPointer
from rsyscall.linux.futex import FutexNode
from rsyscall.sched import Stack, CLONE, _clone, _unshare
from rsyscall.signal import SIG, Siginfo, _kill
from rsyscall.sys.resource import PRIO, _setpriority, _getpriority
from rsyscall.sys.wait import W, ChildState, _waitid
from rsyscall.unistd.credentials import _getpgid, _setpgid
import contextlib
import logging
import rsyscall.far
import rsyscall.near
import typing as t

logger = logging.getLogger(__name__)

@dataclass
class Process:
    """A reference to an arbitrary process on the system, not necessarily our child.

    This is essentially the equivalent of an arbitrary pid. It's not safe to signal
    arbitrary pids, because a process exit + pid wrap can cause you to signal some other
    unexpected process.

    It's only safe to signal child processes, using the ChildProcess class.  But since
    it's common to want to signal arbitrary pids even despite the danger, we provide this
    convenient class to do so.

    """
    task: rsyscall.far.Task
    near: rsyscall.near.Process

    async def kill(self, sig: SIG) -> None:
        await _kill(self.task.sysif, self.near, sig)

    def _as_process_group(self) -> rsyscall.near.ProcessGroup:
        return rsyscall.near.ProcessGroup(self.near.id)

    async def killpg(self, sig: SIG) -> None:
        await _kill(self.task.sysif, self._as_process_group(), sig)

    async def getpgid(self) -> rsyscall.near.ProcessGroup:
        return (await _getpgid(self.task.sysif, self.near))

    async def setpriority(self, prio: int) -> None:
        return (await _setpriority(self.task.sysif, PRIO.PROCESS, self.near.id, prio))

    async def getpriority(self) -> int:
        return (await _getpriority(self.task.sysif, PRIO.PROCESS, self.near.id))

    def __repr__(self) -> str:
        name = type(self).__name__
        return f"{name}({self.near}, parent={self.task})"

class ChildProcess(Process):
    """A process that is our child, which we can monitor with waitid and safely signal.

    Because a child process's pid will not be reused until we wait on its zombie, we can
    (as long as we're careful about ordering calls to waitid and kill) safely send signals
    to child processes without the possibility of signaling some other unexpected process.

    """
    def __init__(self, task: rsyscall.far.Task, near: rsyscall.near.Process, alive=True) -> None:
        self.task = task
        self.near = near
        self.death_state: t.Optional[ChildState] = None
        self.unread_siginfo: t.Optional[Pointer[Siginfo]] = None
        self.in_use = False
        # the command this process exec'd; primarily useful for debugging, but we justify
        # its presence by thinking that the Command might hold references to resources
        # that the process is now using, like temporary directory paths.
        self.command: t.Optional[Command] = None

    def mark_dead(self, state: ChildState) -> None:
        self.death_state = state

    def did_exec(self, command: t.Optional[Command]) -> ChildProcess:
        self.command = command
        return self

    @contextlib.contextmanager
    def borrow(self) -> t.Iterator[None]:
        if self.death_state:
            raise Exception("child process", self.near, "is no longer alive, so we can't wait on it or kill it")
        if self.unread_siginfo:
            raise Exception("for child process", self.near, "waitid or kill was call "
                            "before processing the siginfo buffer from an earlier waitid")
        if self.in_use:
            # TODO technically we could have multiple kills happening simultaneously.
            # but indeed, we can't have a kill happen while a wait is happening, nor multiple waits at a time.
            # that would be racy - we might kill the wrong process or wait on the wrong process
            raise Exception("child process", self.near, "is currently being waited on or killed,"
                            " can't use it a second time")
        self.in_use = True
        try:
            yield
        finally:
            self.in_use = False

    async def kill(self, sig: SIG) -> None:
        with self.borrow():
            await super().kill(sig)

    async def killpg(self, sig: SIG) -> None:
        # This call will throw an error if this child isn't a process group leader, but
        # it's at least guaranteed to not kill some random unrelated process group.
        with self.borrow():
            await super().killpg(sig)

    async def getpgid(self) -> rsyscall.near.ProcessGroup:
        with self.borrow():
            return await super().getpgid()

    async def setpgid(self, pgid: t.Optional[ChildProcess]) -> None:
        # the ownership model of process groups is such that the only way that
        # it's safe to use setpgid on a child process is if we're setpgid-ing to
        # the process group of another child process.
        with self.borrow():
            if pgid is None:
                await _setpgid(self.task.sysif, self.near, None)
            else:
                if pgid.task.pidns != self.task.pidns:
                    raise rsyscall.far.NamespaceMismatchError(
                        "different pid namespaces", pgid.task.pidns, self.task.pidns)
                with pgid.borrow():
                    await _setpgid(self.task.sysif, self.near, self._as_process_group())

    async def waitid(self, options: W, infop: Pointer[Siginfo],
                     *, rusage: t.Optional[Pointer[Siginfo]]=None) -> None:
        with contextlib.ExitStack() as stack:
            stack.enter_context(self.borrow())
            stack.enter_context(infop.borrow(self.task))
            if rusage is not None:
                stack.enter_context(rusage.borrow(self.task))
            try:
                await _waitid(self.task.sysif, self.near, infop.near, options,
                              rusage.near if rusage else None)
            except ChildProcessError as exn:
                exn.filename = self.near
                raise
        self.unread_siginfo = infop

    def parse_waitid_siginfo(self, siginfo: Siginfo) -> t.Optional[ChildState]:
        self.unread_siginfo = None
        # this is to catch the case where we did waitid(W.NOHANG) and there was no event
        if siginfo.pid == 0:
            return None
        else:
            state = ChildState.make_from_siginfo(siginfo)
            if state.died():
                self.mark_dead(state)
            return state

    # helpers
    async def read_siginfo(self) -> t.Optional[ChildState]:
        if self.unread_siginfo is None:
            raise Exception("no siginfo buf to read")
        else:
            siginfo = await self.unread_siginfo.read()
            return self.parse_waitid_siginfo(siginfo)

    async def read_state_change(self) -> ChildState:
        state = await self.read_siginfo()
        if state is None:
            raise Exception("expected a state change, but siginfo buf didn't contain one")
        return state

class ThreadProcess(ChildProcess):
    """A child process with some additional stuff, just useful for resource tracking for threads.

    We need to free the resources used by our child processes when they die. This class
    makes that more straightforward.

    """
    def __init__(self, task: rsyscall.far.Task, near: rsyscall.near.Process,
                 used_stack: Pointer[Stack],
                 stack_data: Stack,
                 ctid: t.Optional[Pointer[FutexNode]],
                 tls: t.Optional[Pointer],
    ) -> None:
        super().__init__(task, near)
        self.used_stack = used_stack
        self.stack_data = stack_data
        self.ctid = ctid
        self.tls = tls

    def free_everything(self) -> None:
        # TODO don't know how to free the stack data...
        if self.used_stack.valid:
            self.used_stack.free()
        if self.ctid is not None and self.ctid.valid:
            self.ctid.free()
        if self.tls is not None and self.tls.valid:
            self.tls.free()

    def mark_dead(self, event: ChildState) -> None:
        self.free_everything()
        return super().mark_dead(event)

    def did_exec(self, command: t.Optional[Command]) -> ChildProcess:
        self.free_everything()
        return super().did_exec(command)

    def __repr__(self) -> str:
        name = type(self).__name__
        if self.command or self.death_state:
            # pretend to be a ChildProcess if we've exec'd or died
            name = 'ChildProcess'
        return f"{name}({self.near}, parent={self.task})"

class ProcessTask(rsyscall.far.Task):
    def __init__(self,
                 sysif: rsyscall.near.SyscallInterface,
                 process: t.Union[rsyscall.near.Process, Process],
                 fd_table: rsyscall.far.FDTable,
                 address_space: rsyscall.far.AddressSpace,
                 pidns: rsyscall.far.PidNamespace,
    ) -> None:
        if isinstance(process, Process):
            near_process = process.near
            self.process = process
            self.parent_task: t.Optional[rsyscall.far.Task] = process.task
        else:
            near_process = process
            self.process = Process(self, process)
            self.parent_task = None
        super().__init__(sysif, near_process, fd_table, address_space, pidns)

    async def clone(self, flags: CLONE,
                    # these two pointers must be adjacent; the end of the first is the start of the
                    # second. the first is the allocation for stack growth, the second is the data
                    # we've written on the stack that will be popped off for arguments.
                    child_stack: t.Tuple[Pointer[Stack], WrittenPointer[Stack]],
                    ptid: t.Optional[Pointer],
                    ctid: t.Optional[Pointer[FutexNode]],
                    # this points to anything, it depends on the thread implementation
                    newtls: t.Optional[Pointer]) -> ThreadProcess:
        clone_parent = bool(flags & CLONE.PARENT)
        if clone_parent:
            if self.parent_task is None:
                raise Exception("using CLONE.PARENT, but we don't know our parent task")
            # TODO also check that the parent_task hasn't shut down... not sure how to do that
            owning_task = self.parent_task
        else:
            owning_task = self
        with contextlib.ExitStack() as stack:
            stack_alloc, stack_data = child_stack
            if (int(stack_data.near) % 16) != 0:
                raise Exception("child stack must have 16-byte alignment, so says Intel")
            stack_alloc_end = stack_alloc.near + stack_alloc.size()
            if stack_alloc_end != stack_data.near:
                raise Exception("the end of the stack allocation pointer", stack_alloc_end,
                                "and the beginning of the stack data pointer", stack_data.near,
                                "must be the same")
            stack.enter_context(stack_alloc.borrow(self))
            stack.enter_context(stack_data.borrow(self))
            ptid_n = self._borrow_optional(stack, ptid)
            ctid_n = self._borrow_optional(stack, ctid)
            newtls_n = self._borrow_optional(stack, newtls)
            process = await _clone(self.sysif, flags, stack_data.near, ptid_n,
                                                ctid_n + ffi.offsetof('struct futex_node', 'futex') if ctid_n else None,
                                                newtls_n)
        # TODO the safety of this depends on no-one borrowing/freeing the stack in borrow __aexit__
        # should try to do this a bit more robustly...
        merged_stack = stack_alloc.merge(stack_data)
        return ThreadProcess(owning_task, process, merged_stack, stack_data.value, ctid, newtls)

    def _make_process(self, pid: int) -> Process:
        return Process(self, rsyscall.near.Process(pid))



"""Allocators and allocations of memory.

None of this is actually specific to memory; it would be a good project to make the
allocators indifferent to whether they are allocating memory or allocating something else
- sub-ranges of files, for example.

It would also be nice to provide an allocator that can grow its memory mapping when it
needs new space. That would require us to pin allocations when making use of pointers
using them.

"""
from __future__ import annotations
from dneio import RequestQueue, reset
from rsyscall._raw import ffi, lib # type: ignore
from rsyscall.far import AddressSpace
from rsyscall.near.sysif import SyscallInterface
import outcome
import rsyscall.far as far
import rsyscall.handle as handle
from rsyscall.memory.allocation_interface import AllocationInterface, UseAfterFreeError
from rsyscall.handle import Task
import trio
import abc
import enum
import contextlib
import typing as t
import logging
from dataclasses import dataclass
from rsyscall.sys.mman import PROT, MAP, MemoryMapping
logger = logging.getLogger(__name__)

# We set eq=False because two distinct zero-length allocations can be identical in all
# their fields, yet they should not be treated as equal, such as in calls to .index()
@dataclass(eq=False)
class Allocation(AllocationInterface):
    """An allocation from some Arena.

    We have a reference back to our Arena so that we can do free(), split(), and merge().
    When __del__ is called on this allocation, we'll free ourselves out of our Arena.

    See AllocationInterface for more about this interface.

    """
    arena: Arena
    start: int
    end: int
    valid: bool = True

    def offset(self) -> int:
        if not self.valid:
            try:
                idx = self.arena.allocations.index(self)
            except ValueError:
                idx = -1
            raise UseAfterFreeError(
                "This allocation has already been freed; refusing to return its offset for use in pointers",
                self,
                "idx", idx,
                "self.arena.allocations", self.arena.allocations,
            )
        return self.start

    def free(self) -> None:
        if self.valid:
            self.valid = False
            self.arena.allocations.remove(self)

    def size(self) -> int:
        return self.end - self.start

    def split(self, size: int) -> t.Tuple[Allocation, Allocation]:
        if not self.valid:
            raise Exception("can't split freed allocation")
        idx = self.arena.allocations.index(self)
        splitpoint = self.start+size
        first = Allocation(self.arena, self.start, splitpoint, valid=False)
        second = Allocation(self.arena, splitpoint, self.end, valid=False)
        self.free()
        self.arena.allocations[idx:idx] = [first, second]
        first.valid = True
        second.valid = True
        return first, second

    def merge(self, other: AllocationInterface) -> Allocation:
        if not isinstance(other, Allocation):
            raise Exception("can't merge", type(self), "with", type(other))
        if not self.valid:
            raise Exception("self.merge(other) was called when self is already freed")
        if not other.valid:
            raise Exception("self.merge(other) was called when other is already freed")
        if self.arena != other.arena:
            # in general, merge is only supported if they started out from the same allocation
            raise Exception("merging allocations from two different arenas - not supported!")
        arena = self.arena
        if self.end != other.start:
            raise Exception("to merge allocations, our end", self.end, "must equal their start", other.start)
        a_idx = arena.allocations.index(self)
        b_idx = arena.allocations.index(other)
        if a_idx + 1 != b_idx:
            raise Exception("allocations are unexpectedly at non-adjacent indices",
                            a_idx, self, id(self), b_idx, other, id(other))
        new = Allocation(self.arena, self.start, other.end, valid=False)
        self.free()
        other.free()
        arena.allocations.insert(a_idx, new)
        new.valid = True
        return new

    def __str__(self) -> str:
        if self.valid:
            return f"Alloc({str(self.arena)}, {self.start}, {self.end})"
        else:
            return f"Alloc(FREED, {str(self.arena)}, {self.start}, {self.end})"

    def __repr__(self) -> str:
        return str(self)

    def __del__(self) -> None:
        # TODO this is actually not going to work, because the Arena stores references to the allocation
        self.free()

class OutOfSpaceError(Exception):
    "Raised by malloc if the allocation request couldn't be satisfied."
    pass

class AllocatorInterface:
    "A memory allocator; raises OutOfSpaceError if there's no more space."
    async def bulk_malloc(self, sizes: t.List[t.Tuple[int, int]]) -> t.Sequence[t.Tuple[MemoryMapping, AllocationInterface]]:
        # A naive bulk allocator
        return [await self.malloc(size, alignment) for size, alignment in sizes]

    @abc.abstractmethod
    async def malloc(self, size: int, alignment: int) -> t.Tuple[MemoryMapping, AllocationInterface]: ...

    def inherit(self, task: Task) -> AllocatorInterface:
        raise Exception("can't be inherited:", self)

@dataclass(eq=False)
class Arena(AllocatorInterface):
    "A single memory mapping and allocations within it."
    mapping: MemoryMapping
    allocations: t.List[Allocation]

    def __init__(self, mapping: MemoryMapping) -> None:
        self.mapping = mapping
        self.allocations: t.List[Allocation] = []

    async def malloc(self, size: int, alignment: int) -> t.Tuple[MemoryMapping, Allocation]:
        return self.mapping, self.allocate(size, alignment)

    def allocate(self, size: int, alignment: int) -> Allocation:
        newstart = 0
        for i, alloc in enumerate(self.allocations):
            if (newstart+size) <= alloc.start:
                newalloc = Allocation(self, newstart, newstart+size)
                self.allocations.insert(i, newalloc)
                return newalloc
            newstart = align(alloc.end, alignment)
        if (newstart+size) <= self.mapping.near.length:
            newalloc = Allocation(self, newstart, newstart+size)
            self.allocations.append(newalloc)
            return newalloc
        raise OutOfSpaceError()

    async def close(self) -> None:
        if self.allocations:
            raise Exception
        await self.mapping.munmap()

    def __str__(self) -> str:
        if len(self.allocations) < 10:
            allocations = "[" + ",".join(f"({alloc.start}, {alloc.end})" for alloc in self.allocations) + "]"
        else:
            allocations = f"[...{len(self.allocations)}...]"
        return f"Arena({str(self.mapping)}, {allocations})"

    def __repr__(self) -> str:
        return str(self)

def align(num: int, alignment: int) -> int:
    """Return the lowest value greater than `num` that is cleanly divisible by `alignment`.

    When applied to an address, this returns the next address that is aligned to this
    alignment. When applied to a size, this returns a size that is sufficient to produce
    an aligned value no matter what its starting address is.

    """
    # TODO this is ugly, isn't there an easier way to do this?
    # we do it this way so that we don't overallocate when overhang is 0.
    overhang = (num % alignment)
    if overhang > 0:
        return num + (alignment - overhang)
    else:
        return num

class UnlimitedAllocator:
    """An allocator which just calls `mmap` to request more memory when it runs out.

    """
    def __init__(self, task: Task) -> None:
        self.task = task
        self.lock = trio.Lock()
        self.arenas: t.List[Arena] = []
        self.queue = RequestQueue[t.List[t.Tuple[int, int]], t.Sequence[t.Tuple[MemoryMapping, Allocation]]]()
        reset(self._run())

    async def _run(self) -> None:
        "Try to allocate all these requests; if we run out of space, make one big mmap call for the rest."
        # TODO we should coalesce together multiple pending mallocs waiting on the lock
        while True:
            sizes, cb = await self.queue.get_one()
            cb.resume(await outcome.acapture(self._bulk_malloc, sizes))

    async def _bulk_malloc(self, sizes: t.List[t.Tuple[int, int]]) -> t.Sequence[t.Tuple[MemoryMapping, Allocation]]:
        "Try to allocate all these requests; if we run out of space, make one big mmap call for the rest."
        allocations: t.List[t.Tuple[MemoryMapping, Allocation]] = []
        size_index = 0
        for arena in self.arenas:
            size, alignment = sizes[size_index]
            if alignment > 4096:
                raise Exception("can't handle alignments of more than 4096 bytes", alignment)
            try:
                allocations.append((arena.mapping, arena.allocate(size, alignment)))
            except OutOfSpaceError:
                pass
            else:
                size_index += 1
                if size_index == len(sizes):
                    # we finished all the allocations, stop allocating
                    break
        if size_index != len(sizes):
            # we hit the end of the arena and now need to allocate more for the remaining sizes:
            rest_sizes = sizes[size_index:]
            # let's do it in bulk:
            # TODO this usage of align() overestimates how much memory we need;
            # it's not a big deal though, because most things have alignment=1
            remaining_size = sum([align(size, alignment) for size, alignment in rest_sizes])
            mapping = await self.task.mmap(align(remaining_size, 4096), PROT.READ|PROT.WRITE, MAP.SHARED)
            arena = Arena(mapping)
            self.arenas.append(arena)
            for size, alignment in rest_sizes:
                if alignment > 4096:
                    raise Exception("can't handle alignments of more than 4096 bytes", alignment)
                try:
                    allocations.append((arena.mapping, arena.allocate(size, alignment)))
                except OutOfSpaceError:
                    raise Exception("some kind of internal error caused a freshly created memory arena",
                                    " to return null for an allocation, size", size, "alignment", alignment)
        return allocations

    async def bulk_malloc(self, sizes: t.List[t.Tuple[int, int]]) -> t.Sequence[t.Tuple[MemoryMapping, Allocation]]:
        return await self.queue.request(sizes)

    async def malloc(self, size: int, alignment: int) -> t.Tuple[MemoryMapping, Allocation]:
        [ret] = await self.bulk_malloc([(size, alignment)])
        return ret

    async def close(self) -> None:
        "Unmap all the mappings owned by this Allocator."
        for arena in self.arenas:
            await arena.close()

class AllocatorClient(AllocatorInterface):
    """A task-specific allocator, to protect us from getting memory back for the wrong address space.

    We share a single allocator between multiple AllocatorClients; we call inherit to make
    a new AllocatorClient. Before returning an allocation, we switch the memory mapping
    handle to be owned by self.task. This checks that the task's address space and the
    mapping's address space match, and ensures that the ownership for the mapping is
    correct.

    """
    def __init__(self, task: Task, shared_allocator: UnlimitedAllocator) -> None:
        self.task = task
        self.shared_allocator = shared_allocator
        if self.task.address_space != self.shared_allocator.task.address_space:
            raise Exception("task and allocator are in different address spaces",
                            self.task.address_space, self.shared_allocator.task.address_space)

    @staticmethod
    def make_allocator(task: Task) -> AllocatorClient:
        return AllocatorClient(task, UnlimitedAllocator(task))

    def inherit(self, task: Task) -> AllocatorClient:
        return AllocatorClient(task, self.shared_allocator)

    async def bulk_malloc(self, sizes: t.List[t.Tuple[int, int]]) -> t.Sequence[t.Tuple[MemoryMapping, AllocationInterface]]:
        seq = await self.shared_allocator.bulk_malloc(sizes)
        return [(mapping.for_task(self.task), alloc) for mapping, alloc in seq]

    async def malloc(self, size: int, alignment: int) -> t.Tuple[MemoryMapping, AllocationInterface]:
        [ret] = await self.bulk_malloc([(size, alignment)])
        return ret



"A trio-enabled variant of unittest.TestCase"
import trio
import unittest
import functools
import types
from trio._core._run import Nursery

class TrioTestCase(unittest.TestCase):
    "A trio-enabled variant of unittest.TestCase"
    nursery: Nursery

    async def asyncSetUp(self) -> None:
        "Asynchronously set up resources for tests in this TestCase"
        pass

    async def asyncTearDown(self) -> None:
        "Asynchronously clean up resources for tests in this TestCase"
        pass

    def __init__(self, methodName='runTest') -> None:
        test = getattr(type(self), methodName)
        @functools.wraps(test)
        async def test_with_setup() -> None:
            async with trio.open_nursery() as nursery:
                self.nursery = nursery
                await self.asyncSetUp()
                try:
                    await test(self)
                except BaseException as exn:
                    try:
                        await self.asyncTearDown()
                    except BaseException as teardown_exn:
                        # have to merge the exceptions if they both throw;
                        # might as well do this with trio.MultiError since we have it
                        raise trio.MultiError([exn, teardown_exn])
                    else:
                        raise
                else:
                    await self.asyncTearDown()
                nursery.cancel_scope.cancel()
        @functools.wraps(test_with_setup)
        def sync_test_with_setup(self) -> None:
            trio.run(test_with_setup)
        setattr(self, methodName, types.MethodType(sync_test_with_setup, self))
        super().__init__(methodName)




"""Convert `help()` output to a string.

`help()` usually sends its output to stdout, which might not be where
we want to write it, if our REPL is targeted somewhere else...

"""

import typing as t
import pydoc # type: ignore

__all__ = [
  'help_to_str',
]

class Output:
  def __init__(self) -> None:
    self.results: t.List[str] = []

  def write(self, s):
    self.results.append(s)

def help_to_str(request: t.Any) -> str:
    "Call `help` on `result`, and return the result as a string"
    out = Output()
    pydoc.Helper(None, out).help(request) # type: ignore
    return "".join(out.results)



"""Various thread implementations with various different special abilities.

Note that all thread types can launch any other thread type, in arbitrary combinations.

Note also that users are able to implement their own types of threads,
so if these available thread types do not meet your needs,
you can implement a new kind of thread which does.

"""



import typing as t

#### Classes ####
from rsyscall.handle.pointer import WrittenPointer
from rsyscall.handle.fd import BaseFileDescriptor, FileDescriptorTask
import os

class CWDTask(FileDescriptorTask):
    async def chdir(self, path: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with path.borrow(self) as path_n:
            await _chdir(self.sysif, path_n)

    async def fchdir(self, fd: BaseFileDescriptor) -> None:
        with fd.borrow(self) as fd_n:
            await _fchdir(self.sysif, fd_n)

    async def chroot(self, path: WrittenPointer[t.Union[str, os.PathLike]]) -> None:
        with path.borrow(self) as path_n:
            await _chroot(self.sysif, path_n)

#### Raw syscalls ####
import rsyscall.near.types as near
from rsyscall.near.sysif import SyscallInterface
from rsyscall.sys.syscall import SYS

async def _chdir(sysif: SyscallInterface, path: near.Address) -> None:
    await sysif.syscall(SYS.chdir, path)

async def _fchdir(sysif: SyscallInterface, fd: near.FileDescriptor) -> None:
    await sysif.syscall(SYS.fchdir, fd)

async def _chroot(sysif: SyscallInterface, path: near.Address) -> None:
    await sysif.syscall(SYS.chroot, path)



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread
import shlex

from rsyscall.sched import CLONE
from rsyscall.fcntl import O
from rsyscall.unistd import SEEK
from rsyscall.stdlib import mkdtemp
from rsyscall.sys.mount import MS
from rsyscall.sys.stat import Stat

from rsyscall.linux.fuse import (
    FuseInitOp, FuseLookupOp, FuseOpenOp, FuseOpendirOp, FuseReadOp, FuseGetattrOp, FuseReaddirplusOp, FuseFlushOp, FuseReleaseOp, FuseReleasedirOp, FuseReadlinkOp, FuseGetxattrOp,
    FuseIn, FuseInList, FuseInitIn, FuseInitOut, FuseAttr, FuseEntryOut, FuseOpenOut, FOPEN,
    FuseAttrOut,
    FuseDirentplus, FuseDirent,
    FUSE_INIT,
)
from rsyscall.linux.dirent import DirentList, DT
from rsyscall.time import Timespec
from rsyscall.sys.stat import TypeMode, S_IF, Mode
from rsyscall import Command, WrittenPointer, FileDescriptor
from rsyscall.memory.ram import RAM
from rsyscall.unistd import AT, ArgList
from rsyscall.tasks.stub import StubServer
from rsyscall.scripts.symsh import FuseFS
import errno
import os
import trio
import typing as t
from dataclasses import dataclass
import unittest

class TestFUSE(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.tmpdir = await mkdtemp(local_thread)
        self.thr = await local_thread.clone(CLONE.NEWUSER|CLONE.NEWNS)
        self.child = await self.thr.clone()
        self.path = self.tmpdir/"path"
        await self.thr.mkdir(self.path)
        self.fuse = await FuseFS.mount(self.thr, self.path)

    async def asyncTearDown(self) -> None:
        await self.fuse.thr.process.kill()
        await self.fuse.cleanup()
        await self.tmpdir.cleanup()

    T_fusein = t.TypeVar('T_fusein', bound=FuseIn)
    async def assertRead(self, cls: t.Type[T_fusein]) -> T_fusein:
        [op] = await self.fuse.read()
        if not isinstance(op, cls):
            raise Exception("expected", cls, "got", op)
        return op

    async def test_basic(self) -> None:
        data_read_from_fuse = b"this is some data read from fuse"
        @self.nursery.start_soon
        async def open() -> None:
            foo = await self.child.task.open(await self.child.ptr(self.path/"foo"), O.RDONLY)
            data, _ = await foo.read(await self.child.malloc(bytes, 4096))
            self.assertEqual(data_read_from_fuse, await data.read())
            data, _ = await foo.read(await self.child.malloc(bytes, 4096))
            self.assertEqual(data.size(), 0)
            await foo.close()

            root = await self.child.task.open(await self.child.ptr(self.path), O.RDONLY)
            valid, rest = await root.getdents(await self.child.ram.malloc(DirentList, 4096))
            await root.close()
        root_getattr = await self.assertRead(FuseGetattrOp)
        if root_getattr.hdr.nodeid != 1:
            raise Exception("expected to get getattr for root node 1,  not", root_getattr.hdr.nodeid)
        await self.fuse.write(root_getattr.respond(FuseAttrOut(
            attr_valid=Timespec(10000, 0),
            attr=FuseAttr(
                ino=1, size=0, blocks=1,
                atime=Timespec(0, 0), mtime=Timespec(0, 0), ctime=Timespec(0, 0),
                mode=TypeMode(S_IF.DIR, Mode(0o777)), nlink=1,
                uid=self.fuse.uid, gid=self.fuse.gid, rdev=0, blksize=4096
            ))))
        await self.fuse.write((await self.assertRead(FuseLookupOp)).respond(FuseEntryOut(
            nodeid=2, generation=1,
            entry_valid=Timespec(10000, 0), attr_valid=Timespec(10000, 0),
            # the size needs to be consistent with the data we'll actually send back on read
            # the kernel, I guess, handles delivering an eof
            attr=FuseAttr(
                ino=999, size=len(data_read_from_fuse), blocks=1,
                atime=Timespec(0, 0), mtime=Timespec(0, 0), ctime=Timespec(0, 0),
                mode=TypeMode(S_IF.REG, Mode(0o777)), nlink=1, uid=self.fuse.uid, gid=self.fuse.gid, rdev=0, blksize=4096
            ))))
        fh = 42
        await self.fuse.write((await self.assertRead(FuseOpenOp)).respond(FuseOpenOut(fh=fh, open_flags=FOPEN.NONE)))
        await self.fuse.write((await self.assertRead(FuseReadOp)).respond(data_read_from_fuse))
        # close file
        await self.fuse.write((await self.assertRead(FuseFlushOp)).respond())
        await self.fuse.write((await self.assertRead(FuseReleaseOp)).respond())
        # open root and getdents
        root_fh = 137
        await self.fuse.write((await self.assertRead(FuseOpendirOp)).respond(FuseOpenOut(fh=root_fh, open_flags=FOPEN.NONE)))
        foobar_ino = 432
        await self.fuse.write((await self.assertRead(FuseReaddirplusOp)).respond([
            FuseDirentplus(
                FuseEntryOut(
                    nodeid=foobar_ino, generation=1,
                    entry_valid=Timespec(10000, 0), attr_valid=Timespec(10000, 0),
                    # the size needs to be consistent with the data we'll actually send back on read
                    # the kernel, I guess, handles delivering an eof
                    attr=FuseAttr(
                        ino=foobar_ino, size=len(data_read_from_fuse), blocks=1,
                        atime=Timespec(0, 0), mtime=Timespec(0, 0), ctime=Timespec(0, 0),
                        mode=TypeMode(S_IF.REG, Mode(0o777)), nlink=1, uid=self.fuse.uid, gid=self.fuse.gid, rdev=0, blksize=4096
                    )),
                FuseDirent(
                    ino=foobar_ino,
                    off=1,
                    type=DT.REG,
                    name="foobar",
                ),
            ),
        ]))
        # close file
        await self.fuse.write((await self.assertRead(FuseReleasedirOp)).respond())

    async def test_symlink(self) -> None:
        @self.nursery.start_soon
        async def open() -> None:
            async with (await self.child.task.open(await self.thr.ptr(self.path/"foo"), O.RDONLY)) as foo:
                data, _ = await foo.read(await self.thr.malloc(bytes, 4096))
        await self.fuse.write((await self.assertRead(FuseGetattrOp)).respond(FuseAttrOut(
            attr_valid=Timespec(10000, 0),
            attr=FuseAttr(
                ino=1, size=0, blocks=1,
                atime=Timespec(0, 0), mtime=Timespec(0, 0), ctime=Timespec(0, 0),
                mode=TypeMode(S_IF.DIR, Mode(0o777)), nlink=1,
                uid=self.fuse.uid, gid=self.fuse.gid, rdev=0, blksize=4096
            ))))
        await self.fuse.write((await self.assertRead(FuseLookupOp)).respond(FuseEntryOut(
            nodeid=2, generation=1,
            entry_valid=Timespec(10000, 0), attr_valid=Timespec(10000, 0),
            # the size needs to be consistent with the data we'll actually send back on read
            # the kernel, I guess, handles delivering an eof;
            # we can just claim a larger size then send back less data
            attr=FuseAttr(
                ino=999, size=4096, blocks=1,
                atime=Timespec(0, 0), mtime=Timespec(0, 0), ctime=Timespec(0, 0),
                mode=TypeMode(S_IF.LNK, Mode(0o777)), nlink=1,
                uid=self.fuse.uid, gid=self.fuse.gid, rdev=0, blksize=4096
            ))))
        await self.fuse.write((await self.assertRead(FuseReadlinkOp)).respond("/bin/sh"))



from rsyscall.tests.trio_test_case import TrioTestCase
from rsyscall import local_thread

from rsyscall.sys.socket import *
from rsyscall.netinet.in_ import *
from rsyscall import Thread, Pointer, FileDescriptor
import trio
import unittest

class TestIP(TrioTestCase):
    async def asyncSetUp(self) -> None:
        self.thr = local_thread

    async def test_stream_listen(self) -> None:
        sockfd = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        addr = await self.thr.bind_getsockname(sockfd, SockaddrIn(0, '127.0.0.1'))
        await sockfd.listen(10)

        real_addr = await self.thr.ram.ptr(addr)
        clientfd = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await clientfd.connect(real_addr)
        connfd = await sockfd.accept()

        in_data = await self.thr.ram.ptr(b"hello")
        written, _ = await clientfd.write(in_data)
        valid, _ = await connfd.read(written)
        self.assertEqual(in_data.value, await valid.read())

    async def test_dgram_connect(self) -> None:
        sockfd = await self.thr.task.socket(AF.INET, SOCK.DGRAM)
        addr = await self.thr.bind_getsockname(sockfd, SockaddrIn(0, '127.0.0.1'))

        real_addr = await self.thr.ram.ptr(addr)
        clientfd = await self.thr.task.socket(AF.INET, SOCK.DGRAM)
        await clientfd.connect(real_addr)

        in_data = await self.thr.ram.ptr(b"hello")
        written, _ = await clientfd.write(in_data)
        valid, _ = await sockfd.read(written)
        self.assertEqual(in_data.value, await valid.read())

    async def test_write_to_unconnected(self) -> None:
        sockfd = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        with self.assertRaises(BrokenPipeError):
            await sockfd.write(await self.thr.ram.ptr(b"hello"))

    @unittest.skip("This test is slow and non-deterministic")
    async def test_send_is_not_atomic(self) -> None:
        """send does, in fact, do partial writes, at least when set to NONBLOCK

        That is to say, it won't return anything less than sending the whole pointer, or (in the
        case of NONBLOCK) sending none of the pointer. This is supported by the manpage, which says:
        "When the message does not fit into the send buffer of the socket, send() normally blocks".

        Unfortunately, it's still not atomic. If multiple threads are sending at once, the data can
        be interleaved.

        """
        sockfd = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        addr = await self.thr.bind_getsockname(sockfd, SockaddrIn(0, '127.0.0.1'))
        await sockfd.listen(10)

        real_addr = await self.thr.ram.ptr(addr)
        clientfd = await self.thr.task.socket(AF.INET, SOCK.STREAM)
        await clientfd.connect(real_addr)
        connfd = await sockfd.accept()
        orig_in_fd = clientfd
        orig_out_fd = connfd
        data = "".join(str(i) for i in range(8000)).encode()

        count = 100
        threads = [await self.thr.clone() for _ in range(10)]
        in_ptrs = [await thr.ptr(data) for thr in threads]
        handles = [thr.task.inherit_fd(orig_in_fd) for thr in threads]
        async def run_send(thread: Thread, in_ptr: Pointer, fd: FileDescriptor) -> None:
            for i in range(count):
                in_ptr, rest = await fd.write(in_ptr)
                if rest.size() != 0:
                    print("failure! rest.size() is", rest.size())
                self.assertEqual(rest.size(), 0)
        read_thread = await self.thr.clone()
        out_buf = await read_thread.malloc(bytes, len(data))
        out_fd = read_thread.inherit_fd(orig_out_fd)

        had_interleaving = False
        async with trio.open_nursery() as nursery:
            for thread, in_ptr, fd in zip(threads, in_ptrs, handles):
                nursery.start_soon(run_send, thread, in_ptr, fd)
            for i in range(len(threads) * count):
                out_buf, rest = await out_fd.recv(out_buf, MSG.WAITALL)
                self.assertEqual(rest.size(), 0)
                if not had_interleaving:
                    indata = await out_buf.read()
                    if indata != data:
                        # oops, looks like the data from multiple threads was interleaved
                        had_interleaving = True
        for thread in threads:
            await thread.exit(0)
        await read_thread.exit(0)
        self.assertTrue(had_interleaving)



"""Memory transport to a remote address space based on reading/writing file descriptors.

We need to be able to read and write to the address spaces of our
threads. We start with the ability to read and write to the address
space of the local thread; we need to bootstrap that into an ability
to read and write to other address spaces.

We could have taken a dependency on some means of RDMA, or used
process_vm_readv/writev, or used other techniques, but these only work
in certain circumstances and places additional dependencies on
us. Nevertheless, they might be useful optimizations in the future.

Instead, we make sure that whenever we want to read and write to any
address space, we have a file descriptor owned by a task in that
address space, which is connected to a file descriptor owned by a task
in an address space that we already can read and write.

Then, the problem is reduced to copying bytes between memory and the
connection underlying the pair of file descriptors. But that's easy:
We just use the "read" and "write" system calls.

So, to write to memory in some address space B, we write to some
memory in some address space A that we already can access; then call
`write` on A's file descriptor to copy the bytes from memory in A into
the connection; then call `read` on B's file descriptor to copy the
bytes from the connection to memory in B.

To read memory in address space B, we just perform this in reverse; we
call `write` on B's file descriptor, and `read` on A's file descriptor.

"""
from __future__ import annotations
from dataclasses import dataclass
from rsyscall import AsyncFileDescriptor, Pointer, FileDescriptor, Task
from rsyscall.memory.allocation_interface import AllocationInterface
from rsyscall.memory.transport import MemoryTransport
from rsyscall.sys.socket import MSG
import logging
import trio
import typing as t

__all__ = [
    "SocketMemoryTransport",
]

logger = logging.getLogger(__name__)

@dataclass
class SpanAllocation(AllocationInterface):
    """An allocation which is a subspan of some other allocation, and can be split freely

    This should be built into our allocation system. In fact, it is: This is what split is
    for. But the ownership is tricky: Splitting an allocation consumes it. We aren't
    supposed to take ownership of the pointers passed to us for write/read, so
    we can't naively split the pointers.  Instead, we use to_span, below, to make them use
    SpanAllocation, so we can split them freely without taking ownership.

    We should make it possible to split an allocation without consuming it, or otherwise
    have multiple references to the same allocation, then we can get rid of this.

    """
    alloc: AllocationInterface
    _offset: int
    _size: int

    def __post_init__(self) -> None:
        if self._offset + self._size > self.alloc.size():
            raise Exception("span falls off the end of the underlying allocation",
                            self._offset, self._size, self.alloc.size())

    def offset(self) -> int:
        return self.alloc.offset() + self._offset

    def size(self) -> int:
        return self._size

    def split(self, size: int) -> t.Tuple[AllocationInterface, AllocationInterface]:
        if size > self.size():
            raise Exception("called split with size", size, "greater than this allocation's total size", self.size())
        return (SpanAllocation(self.alloc, self._offset, size),
                SpanAllocation(self.alloc, self._offset + size, self._size - size))

    def merge(self, other: AllocationInterface) -> AllocationInterface:
        if not isinstance(other, SpanAllocation):
            raise Exception("can only merge SpanAllocation with SpanAllocation, not", other)
        if self.alloc == other.alloc:
            if self._offset + self._size == other._offset:
                return SpanAllocation(self.alloc, self._offset, self._size + other._size)
            else:
                raise Exception("spans are not adjacent")
        else:
            raise Exception("can't merge spans over two different allocations")

    def free(self) -> None:
        pass

def to_span(ptr: Pointer) -> Pointer:
    "Wraps the pointer's allocation in SpanAllocation so it can be split freely"
    return Pointer(
        ptr.mapping,
        ptr.transport,
        ptr.serializer,
        SpanAllocation(ptr.allocation, 0, ptr.allocation.size()),
        ptr.typ)

@dataclass
class SocketMemoryTransport(MemoryTransport):
    """Read and write bytes from a remote address space, using a connected socketpair

    We use the "ram" inside the "local" AsyncFileDescriptor to access
    pointers in the address space of the "local" fd. We use the
    socketpair to transport memory from the "local" address space to
    the address space of the "remote" fd.

    In this way, we turn a transport for the "local" address space,
    plus the connected socketpair, into a transport for the "remote"
    address space.

    We rely heavily on the fact that syscall results are returned
    "synchronously" and in order. After a socket read or write completes on one
    end of the socketpair, we rely on the corresponding write or read on the
    other end of the socketpair to be submitted immediately, so that the data is
    processed in order.

    """
    local: AsyncFileDescriptor
    remote: FileDescriptor

    def inherit(self, task: Task) -> SocketMemoryTransport:
        return SocketMemoryTransport(self.local, self.remote.for_task(task))

    async def do_write(self, dest: Pointer, src: Pointer[bytes], i=0) -> None:
        if dest.size() == 0:
            return
        written, rest = await self.local.write(src)
        dest, dest_rest = dest.split(written.size())
        if dest_rest.size() != 0:
            raise NotImplementedError("partial write, oops, not supported yet")
        await self.remote.recv(dest, MSG.WAITALL)

    async def write(self, dest: Pointer, data: bytes) -> None:
        if dest.size() != len(data):
            raise Exception("mismatched pointer size", dest.size(), "and data size", len(data))
        src = await self.local.ram.ptr(data)
        dest_span = to_span(dest)
        await self.do_write(dest_span, src)

    async def do_read(self, dest: Pointer[bytes], src: Pointer) -> bytes:
        if dest.size() == 0:
            return b''
        written, rest = await self.remote.write(src)
        dest, dest_rest = dest.split(written.size())
        if dest_rest.size() != 0:
            raise NotImplementedError("partial write, oops, not supported yet")
        result, anything_left = await self.local.read(dest)
        if anything_left.size() != 0:
            raise NotImplementedError("partial read, oops, not supported yet")
        return await result.read()

    async def read(self, src: Pointer) -> bytes:
        dest = await self.local.ram.malloc(bytes, src.size())
        src_span = to_span(src)
        return await self.do_read(dest, src_span)
